"""Plot annotations."""

from __future__ import annotations

__all__ = [
    "SI_PREFIXES",
    "SI_PREFIX_NAMES",
    "copy_mathtext",
    "fancy_labels",
    "label_subplot_properties",
    "label_subplots",
    "label_subplots_nature",
    "mark_points",
    "mark_points_outside",
    "plot_hv_text",
    "property_label",
    "scale_units",
    "set_titles",
    "set_xlabels",
    "set_ylabels",
    "sizebar",
]

import io
import re
from typing import TYPE_CHECKING, Literal

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.ticker
import matplotlib.transforms as mtransforms
import numpy as np
import pyperclip
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar

from erlab.plotting.colors import axes_textcolor

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

SI_PREFIXES: dict[int, str] = {
    24: "Y",
    21: "Z",
    18: "E",
    15: "P",
    12: "T",
    9: "G",
    6: "M",
    3: "k",
    2: "h",
    1: "da",
    0: "",
    -1: "d",
    -2: "c",
    -3: "m",
    -6: "μ",
    -9: "n",
    -12: "p",
    -15: "f",
    -18: "a",
    -21: "z",
    -24: "y",
}  #: Maps powers of 10 to valid SI prefix strings.

SI_PREFIX_NAMES: tuple[str, ...] = (
    "yotta",
    "zetta",
    "exa",
    "peta",
    "tera",
    "giga",
    "mega",
    "kilo",
    "hecto",
    "deca",
    "",
    "deci",
    "centi",
    "milli",
    "micro",
    "nano",
    "pico",
    "femto",
    "atto",
    "zepto",
    "yocto",
)  #: Names of the SI prefixes.


def _alph_label(val, prefix, suffix, numeric, capital):
    """Generate labels from string or integer."""
    if isinstance(val, int | np.integer) or val.isdigit():
        if numeric:
            val = str(val)
        else:
            if capital:
                ref_char = "A"
            else:
                ref_char = "a"
            val = chr(int(val) + ord(ref_char) - 1)
    elif not isinstance(val, str):
        raise TypeError("Input values must be integers or strings.")
    return prefix + val + suffix


def _unit_from_label(label: str) -> str | None:
    """Try to determine the unit from a given label.

    Returns None if it fails to determine the unit.

    """
    m = re.match(r".*\((.*)\)\s*?$", label)
    if m is None:
        return None
    try:
        return m.group(1)
    except IndexError:
        return None


def get_si_str(si: int) -> str:
    """Returns the SI prefix string to be plotted by :mod:`matplotlib`.

    Parameters
    ----------
    si : int
        Exponent of 10 corresponding to a SI prefix.

    Returns
    -------
    str
        SI prefix corresponding to ``si``.

    """
    if plt.rcParams["text.usetex"] and si == -6:
        return "\\ensuremath{\\mu}"
    else:
        try:
            return SI_PREFIXES[si]
        except KeyError as e:
            raise ValueError("Invalid SI prefix.") from e


def name_for_dim(dim_name, escaped=True):
    name = {
        "temperature": ("Temperature", "Temperature"),
        "T": (r"\ensuremath{T}", r"$T$"),
        "beta": (r"\ensuremath{\beta}", r"$\beta$"),
        "theta": (r"\ensuremath{\theta}", r"$\theta$"),
        "chi": (r"\ensuremath{\chi}", r"$\chi$"),
        "alpha": (r"\ensuremath{\alpha}", r"$\alpha$"),
        "psi": (r"\ensuremath{\psi}", r"$\psi$"),
        "phi": (r"\ensuremath{\phi}", r"$\phi$"),
        "xi": (r"\ensuremath{\xi}", r"$\xi$"),
        "Eb": (r"\ensuremath{E}", r"$E$"),
        "eV": (r"\ensuremath{E-E_F}", r"$E-E_F$"),
        "kx": (r"\ensuremath{k_{x}}", r"$k_x$"),
        "ky": (r"\ensuremath{k_{y}}", r"$k_y$"),
        "kz": (r"\ensuremath{k_{z}}", r"$k_z$"),
        "kp": (r"\ensuremath{k_{\parallel}}", r"$k_\parallel$"),
        "hv": (r"\ensuremath{h\nu}", r"$h\nu$"),
    }.get(dim_name)

    if name is None:
        name = dim_name
    else:
        name = name[0] if plt.rcParams["text.usetex"] else name[1]

    if not escaped:
        name = name.replace("$", "")
    return name


def unit_for_dim(dim_name, deg2rad=False):
    unit = {
        "temperature": (r"K", r"K"),
        "T": (r"K", r"K"),
        "theta": (r"deg", r"deg"),
        "beta": (r"deg", r"deg"),
        "psi": (r"deg", r"deg"),
        "chi": (r"deg", r"deg"),
        "alpha": (r"deg", r"deg"),
        "phi": (r"deg", r"deg"),
        "xi": (r"deg", r"deg"),
        "Eb": (r"eV", r"eV"),
        "eV": (r"eV", r"eV"),
        "hv": (r"eV", r"eV"),
        "kx": (r"Å\ensuremath{{}^{-1}}", r"Å${}^{-1}$"),
        "ky": (r"Å\ensuremath{{}^{-1}}", r"Å${}^{-1}$"),
        "kz": (r"Å\ensuremath{{}^{-1}}", r"Å${}^{-1}$"),
        "kp": (r"Å\ensuremath{{}^{-1}}", r"Å${}^{-1}$"),
    }.get(dim_name)

    if unit is None:
        unit = ""
    else:
        unit = unit[0] if plt.rcParams["text.usetex"] else unit[1]
    if deg2rad:
        unit = unit.replace("deg", "rad")
    return unit


def label_for_dim(dim_name, deg2rad=False, escaped=True):
    name = name_for_dim(dim_name, escaped=escaped)
    unit = unit_for_dim(dim_name, deg2rad=deg2rad)
    if unit == "":
        return name
    else:
        return f"{name} ({unit})"


def parse_special_point(name):
    special_points = {"G": r"\Gamma", "D": r"\Delta"}
    try:
        return special_points[name]
    except KeyError:
        return name


def parse_point_labels(name: str, roman=True, bar=False):
    name = parse_special_point(name)

    if name.endswith("*"):
        name = name[:-1]
        if roman:
            format_str = r"\mathdefault{{{}}}^*"
        else:
            format_str = r"{}^*"
    elif name.endswith("'"):
        name = name[:-1]
        if roman:
            format_str = r"\mathdefault{{{}}}\prime"
        else:
            format_str = r"{}\prime"
    elif roman:
        format_str = r"\mathdefault{{{}}}"
    else:
        format_str = r"{}"

    name = format_str.format(parse_special_point(name))

    if bar:
        name = rf"$\overline{{{name}}}$"
    else:
        name = rf"${name}$"

    return name


def copy_mathtext(
    s: str,
    fontsize=None,
    fontproperties=None,
    outline=False,
    svg=False,
    rcparams: dict | None = None,
    **mathtext_rc,
):
    if fontproperties is None:
        fontproperties = matplotlib.font_manager.FontProperties(size=fontsize)
    else:
        fontproperties.set_size(fontsize)
    if rcparams is None:
        rcparams = {}
    parser = matplotlib.mathtext.MathTextParser("path")
    width, height, depth, _, _ = parser.parse(s, dpi=72, prop=fontproperties)
    fig = matplotlib.figure.Figure(figsize=(width / 72, height / 72))
    fig.patch.set_facecolor("none")
    fig.text(0, depth / height, s, fontproperties=fontproperties)

    if svg:
        matplotlib.backends.backend_svg.FigureCanvasSVG(fig)
    else:
        matplotlib.backends.backend_pdf.FigureCanvasPdf(fig)

    for k, v in mathtext_rc.items():
        if k in ["bf", "cal", "it", "rm", "sf", "tt"] and isinstance(
            v, matplotlib.font_manager.FontProperties
        ):
            v = v.get_fontconfig_pattern()
        rcparams[f"mathtext.{k}"] = v

    with io.BytesIO() as buffer:
        if svg:
            rcparams.setdefault("svg.fonttype", "path" if outline else "none")
            rcparams.setdefault("svg.image_inline", True)
            with plt.rc_context(rcparams):
                fig.canvas.print_svg(buffer)
        else:
            rcparams.setdefault("pdf.fonttype", 3 if outline else 42)
            with plt.rc_context(rcparams):
                fig.canvas.print_pdf(buffer)
        pyperclip.copy(buffer.getvalue().decode("utf-8"))


def fancy_labels(ax=None, deg2rad=False):
    if ax is None:
        ax = plt.gca()
    if np.iterable(ax):
        for axi in ax:
            fancy_labels(axi, deg2rad)
        return

    ax.set_xlabel(label_for_dim(dim_name=ax.get_xlabel(), deg2rad=deg2rad))
    ax.set_ylabel(label_for_dim(dim_name=ax.get_ylabel(), deg2rad=deg2rad))
    if hasattr(ax, "get_zlabel"):
        ax.set_zlabel(label_for_dim(dim_name=ax.get_zlabel(), deg2rad=deg2rad))


def label_subplot_properties(
    axes: matplotlib.axes.Axes | Sequence[matplotlib.axes.Axes],
    values: dict,
    decimals: int | None = None,
    si: int = 0,
    name: str | None = None,
    unit: str | None = None,
    order: Literal["C", "F", "A", "K"] = "C",
    **kwargs: dict,
):
    r"""Labels subplots with automatically generated labels.

    Parameters
    ----------
    axes
        `matplotlib.axes.Axes` to label. If an array is given, the order will be
        determined by the flattening method given by `order`.
    values
        key-value pair of annotations.
    decimals
        Number of decimal places to round to. If decimals is None, no
        rounding is performed. If decimals is negative, it specifies the
        number of positions to the left of the decimal point.
    si
        Powers of 10 for automatic SI prefix setting.
    name
        When set, overrides automatic dimension name setting.
    unit
        When set, overrides automatic unit setting.
    order
        Order in which to flatten `ax`. 'C' means to flatten in
        row-major (C-style) order. 'F' means to flatten in column-major
        (Fortran- style) order. 'A' means to flatten in column-major
        order if a is Fortran contiguous in memory, row-major order
        otherwise. 'K' means to flatten a in the order the elements
        occur in memory. The default is 'C'.
    **kwargs
        Extra arguments to `erlab.plotting.annotations.label_subplots`.

    """
    kwargs.setdefault("fontweight", plt.rcParams["font.weight"])
    kwargs.setdefault("prefix", "")
    kwargs.setdefault("suffix", "")
    kwargs.setdefault("loc", "upper right")

    strlist = []
    for k, v in values.items():
        if not isinstance(v, tuple | list | np.ndarray):
            v = [v]
        else:
            v = np.array(v).flatten(order=order)
        strlist.append(
            [
                property_label(k, val, decimals=decimals, si=si, name=name, unit=unit)
                for val in v
            ]
        )
    strlist = list(zip(*strlist))
    strlist = ["\n".join(strlist[i]) for i in range(len(strlist))]
    label_subplots(axes, strlist, order=order, **kwargs)


def label_subplots(
    axes: matplotlib.axes.Axes | Sequence[matplotlib.axes.Axes],
    values: Sequence[int | str] | None = None,
    startfrom: int = 1,
    order: Literal["C", "F", "A", "K"] = "C",
    loc: Literal[
        "upper left",
        "upper center",
        "upper right",
        "center left",
        "center",
        "center right",
        "lower left",
        "lower center",
        "lower right",
    ] = "upper left",
    offset: tuple[float, float] = (0.0, 0.0),
    prefix: str = "",
    suffix: str = "",
    numeric: bool = False,
    capital: bool = False,
    fontweight: Literal[
        "ultralight",
        "light",
        "normal",
        "regular",
        "book",
        "medium",
        "roman",
        "semibold",
        "demibold",
        "demi",
        "bold",
        "heavy",
        "extra bold",
        "black",
    ] = "normal",
    fontsize: (
        float
        | Literal[
            "xx-small", "x-small", "small", "medium", "large", "x-large", "xx-large"
        ]
        | None
    ) = None,
    **kwargs,
):
    r"""Labels subplots with automatically generated labels.

    Parameters
    ----------
    axes
        `matplotlib.axes.Axes` to label. If an array is given, the order will be
        determined by the flattening method given by `order`.
    values
        Integer or string labels corresponding to each Axes in `axes` for
        manual labels.
    startfrom
        Start from this number when creating automatic labels. Has no
        effect when `values` is not `None`.
    order
        Order in which to flatten `ax`. 'C' means to flatten in
        row-major (C-style) order. 'F' means to flatten in column-major
        (Fortran- style) order. 'A' means to flatten in column-major
        order if a is Fortran contiguous in memory, row-major order
        otherwise. 'K' means to flatten a in the order the elements
        occur in memory. The default is 'C'.
    loc
        The box location. The default is ``'upper left'``.
    offset
        Values that are used to position the legend in conjunction with
        `loc`, given in display units.
    prefix
        String to prepend to the alphabet label.
    suffix
        String to append to the alphabet label.
    numeric
        Use integer labels instead of alphabets.
    capital
        Capitalize automatically generated alphabetical labels.
    fontweight
        Set the font weight. The default is ``'normal'``.
    fontsize
        Set the font size. The default is ``'medium'`` for axes, and ``'large'`` for
        figures.
    **kwargs
        Extra arguments to `matplotlib.text.Text`: refer to the `matplotlib`
        documentation for a list of all possible arguments.

    """

    kwargs["fontweight"] = fontweight
    if plt.rcParams["text.usetex"] & (fontweight == "bold"):
        prefix = "\\textbf{" + prefix
        suffix = suffix + "}"
        kwargs.pop("fontweight")

    axlist = np.array(axes, dtype=object).flatten(order=order)
    if values is None:
        values = np.array([i + startfrom for i in range(len(axlist))], dtype=np.int64)
    else:
        values = np.array(values).flatten(order=order)
        if not (axlist.size == values.size):
            raise IndexError(
                "The number of given values must match the number of given axes."
            )

    for i in range(len(axlist)):
        bbox_to_anchor = axlist[i].bbox
        if fontsize is None:
            if isinstance(axlist[i], matplotlib.figure.Figure):
                fs = "large"
            else:
                fs = "medium"
        else:
            fs = fontsize

        bbox_transform = matplotlib.transforms.ScaledTranslation(
            offset[0] / 72, offset[1] / 72, axlist[i].get_figure().dpi_scale_trans
        )
        label_str = _alph_label(values[i], prefix, suffix, numeric, capital)
        with plt.rc_context({"text.color": axes_textcolor(axlist[i])}):
            at = matplotlib.offsetbox.AnchoredText(
                label_str,
                loc=loc,
                frameon=False,
                pad=0,
                borderpad=0.5,
                prop=dict(fontsize=fs, **kwargs),
                bbox_to_anchor=bbox_to_anchor,
                bbox_transform=bbox_transform,
                clip_on=False,
            )
        axlist[i].add_artist(at)


def label_subplots_nature(
    axes: matplotlib.axes.Axes | Sequence[matplotlib.axes.Axes],
    values: Sequence[int | str] | None = None,
    startfrom: int = 1,
    order: Literal["C", "F", "A", "K"] = "C",
    offset: tuple[float, float] = (-20.0, 7.0),
    prefix: str = "",
    suffix: str = "",
    numeric: bool = False,
    capital: bool = False,
    fontweight: Literal[
        "ultralight",
        "light",
        "normal",
        "regular",
        "book",
        "medium",
        "roman",
        "semibold",
        "demibold",
        "demi",
        "bold",
        "heavy",
        "extra bold",
        "black",
    ] = "black",
    fontsize: (
        float
        | Literal[
            "xx-small", "x-small", "small", "medium", "large", "x-large", "xx-large"
        ]
    ) = 8,
    **kwargs,
):
    r"""Labels subplots with automatically generated labels.

    Parameters
    ----------
    axes
        `matplotlib.axes.Axes` to label. If an array is given, the order will be
        determined by the flattening method given by `order`.
    values
        Integer or string labels corresponding to each Axes in `axes` for
        manual labels.
    startfrom
        Start from this number when creating automatic labels. Has no
        effect when `values` is not `None`.
    order
        Order in which to flatten `ax`. 'C' means to flatten in
        row-major (C-style) order. 'F' means to flatten in column-major
        (Fortran- style) order. 'A' means to flatten in column-major
        order if a is Fortran contiguous in memory, row-major order
        otherwise. 'K' means to flatten a in the order the elements
        occur in memory. The default is 'C'.
    offset
        Values that are used to position the labels, given in points.
    prefix
        String to prepend to the alphabet label.
    suffix
        String to append to the alphabet label.
    numeric
        Use integer labels instead of alphabets.
    capital
        Capitalize automatically generated alphabetical labels.
    fontweight
        Set the font weight. The default is ``'normal'``.
    fontsize
        Set the font size. The default is ``'medium'`` for axes, and ``'large'`` for
        figures.
    **kwargs
        Extra arguments to `matplotlib.text.Text`: refer to the `matplotlib`
        documentation for a list of all possible arguments.

    """

    kwargs["fontweight"] = fontweight
    if plt.rcParams["text.usetex"] & (fontweight == "bold"):
        prefix = "\\textbf{" + prefix
        suffix = suffix + "}"
        kwargs.pop("fontweight")

    axlist = np.array(axes, dtype=object).flatten(order=order)
    if values is None:
        values = np.array([i + startfrom for i in range(len(axlist))], dtype=np.int64)
    else:
        values = np.array(values).flatten(order=order)
        if not (axlist.size == values.size):
            raise IndexError(
                "The number of given values must match the number of given axes."
            )

    for i in range(len(axlist)):
        label_str = _alph_label(values[i], prefix, suffix, numeric, capital)
        trans = matplotlib.transforms.ScaledTranslation(
            offset[0] / 72, offset[1] / 72, axlist[i].get_figure().dpi_scale_trans
        )

        if fontsize is None:
            fontsize = "medium"
        axlist[i].figure.text(
            # axlist[i].text(
            0.0,
            1.0,
            label_str,
            transform=axlist[i].transAxes + trans,
            fontsize=fontsize,
            va="baseline",
            clip_on=False,
            **kwargs,
        )


def mark_points(
    points: Sequence[float],
    labels: Sequence[str],
    y: float | Sequence[float] = 0.0,
    pad: tuple[float, float] = (0, 1.75),
    literal: bool = False,
    roman: bool = True,
    bar: bool = False,
    ax: matplotlib.axes.Axes | Iterable[matplotlib.axes.Axes] = None,
    **kwargs: dict,
):
    """Mark points above the horizontal axis.

    Useful when annotating high symmetry points along a cut.

    Parameters
    ----------
    points
        Floats indicating the position of each label.
    labels
        Sequence of label strings indicating a high symmetry point. Must be the same
        length as `points`.
    y
        Position of the label in data coordinates
    pad
        Offset of the text in points.
    literal
        If `True`, take the input string literally.
    roman
        If ``False``, *True*, itallic fonts are used.
    bar
        If ``True``, prints a bar over the label.
    ax
        `matplotlib.axes.Axes` to annotate.

    """
    if ax is None:
        ax = plt.gca()
    if np.iterable(ax):
        for a in np.asarray(ax, dtype=object).flatten():
            mark_points(points, labels, y, pad, literal, roman, bar, a, **kwargs)
    else:
        for k, v in {"ha": "center", "va": "baseline", "fontsize": "small"}.items():
            kwargs.setdefault(k, v)
        if not np.iterable(y):
            y = [y] * len(points)
        with plt.rc_context({"font.family": "serif"}):
            for xi, yi, label in zip(points, y, labels):
                ax.text(
                    xi,
                    yi,
                    label if literal else parse_point_labels(label, roman, bar),
                    transform=ax.transData
                    + mtransforms.ScaledTranslation(
                        pad[0] / 72, pad[1] / 72, ax.figure.dpi_scale_trans
                    ),
                    **kwargs,
                )


def mark_points_outside(
    points: Sequence[float],
    labels: Sequence[str],
    axis: Literal["x", "y"] = "x",
    roman: bool = True,
    bar: bool = False,
    ax: matplotlib.axes.Axes | Iterable[matplotlib.axes.Axes] = None,
):
    """Mark points above the horizontal axis.

    Useful when annotating high symmetry points along a cut.

    Parameters
    ----------
    points
        Floats indicating the position of each label.
    labels
        Sequence of label strings indicating a high symmetry point. Must be the same
        length as `points`.
    axis
        If ``'x'``, marks points along the horizontal axis. If ``'y'``, marks points
        along the vertical axis.
    roman
        If ``False``, *True*, itallic fonts are used.
    bar
        If ``True``, prints a bar over the label.
    ax
        `matplotlib.axes.Axes` to annotate.

    """
    if ax is None:
        ax = plt.gca()
    if np.iterable(ax):
        for a in np.asarray(ax, dtype=object).flatten():
            mark_points_outside(points, labels, axis, roman, bar, a)
    else:
        if axis == "x":
            label_ax = ax.twiny()
            label_ax.set_xlim(ax.get_xlim())
            label_ax.set_xticks(points)
            label_ax.set_xticklabels(
                [parse_point_labels(lab, roman, bar) for lab in labels]
            )
        else:
            label_ax = ax.twinx()
            label_ax.set_ylim(ax.get_ylim())
            label_ax.set_yticks(points)
            label_ax.set_yticklabels(
                [parse_point_labels(lab, roman, bar) for lab in labels]
            )
        label_ax.set_frame_on(False)


def mark_points_y(pts, labels, roman=True, bar=False, ax=None):
    if ax is None:
        ax = plt.gca()
    if not isinstance(ax, tuple | list | np.ndarray):
        ax = [ax]
    for a in np.array(ax, dtype=object).flatten():
        label_ax = a.twinx()
        label_ax.set_ylim(a.get_ylim())
        label_ax.set_yticks(pts)
        # label_ax.set_xlabel('')
        label_ax.set_yticklabels(
            [parse_point_labels(lab, roman, bar) for lab in labels]
        )
        # label_ax.set_zorder(a.get_zorder())
        label_ax.set_frame_on(False)


# TODO: fix format using name_for_dim and unit_for_dim
def plot_hv_text(ax, val, x=0.025, y=0.975, **kwargs):
    s = "$h\\nu=" + str(val) + "$~eV"
    ax.text(
        x,
        y,
        s,
        family="serif",
        horizontalalignment="left",
        verticalalignment="top",
        transform=ax.transAxes,
        **kwargs,
    )


def plot_hv_text_right(ax, val, x=1 - 0.025, y=0.975, **kwargs):
    s = "$h\\nu=" + str(val) + "$~eV"
    ax.text(
        x,
        y,
        s,
        family="serif",
        horizontalalignment="right",
        verticalalignment="top",
        transform=ax.transAxes,
        **kwargs,
    )


def property_label(key, value, decimals=None, si=0, name=None, unit=None):
    if name == "":
        delim = ""
    else:
        delim = " = "
    if name is None:
        name = name_for_dim(key, escaped=False)
        if name is None:
            name = ""
    if unit is None:
        unit = unit_for_dim(key)
        if unit is None:
            unit = ""

    unit = get_si_str(si) + unit
    value /= 10**si
    if decimals is not None:
        value = np.around(value, decimals=decimals)
    if int(value) == value:
        value = int(value)

    if key == "Eb":
        if value == 0:
            if delim == "":
                return "$E_F$"
            else:
                return "$E = E_F$"
        if delim == "":
            name = "E_F"
        else:
            delim += "E_F"
        if value > 0:
            delim += "+"

    base = "${}" + delim + "{}$ {}"
    return str(base.format(name, value, unit))


class _SIFormatter(matplotlib.ticker.ScalarFormatter):
    def __init__(self, si: int = 0, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._si_exponent = int(si)

    def __call__(self, x, pos=None):
        self.orderOfMagnitude += self._si_exponent

        sigfigs = int(re.match(r".*{%1.(\d+)f}", self.format).group(1))
        self.format = self.format.replace(
            f"%1.{sigfigs}f", f"%1.{max(0, sigfigs + self._si_exponent)}f"
        )
        val = super().__call__(x, pos)

        self.orderOfMagnitude -= self._si_exponent
        return val


def scale_units(
    ax: matplotlib.axes.Axes,
    axis: Literal["x", "y", "z"],
    si: int = 0,
    *,
    prefix: bool = True,
    power: bool = False,
):
    """Rescales ticks and adds an SI prefix to the axis label.

    Useful when you want to rescale the ticks. For example, when plotting a cut from a
    low pass energy scan, you might want to convert the energy units from eV to meV.

    Using this function on an axis where the major locator is not the default formatter
    `matplotlib.ticker.ScalarFormatter` will result in undefined behavior.

    Parameters
    ----------
    ax
        _description_
    axis
        The axis you wish to rescale.
    si
        Exponent of 10 corresponding to a SI prefix.
    prefix
        If True, tries to detect the unit from the axis label and scales it accordingly.
        The scaling behaviour is controlled by the `power` argument. If no units are
        found in the axis label, it is silently ignored.
    power
        If False, prefixes the detected unit on the axis label with a SI prefix
        corresponding to `si`. If True, the unit is prefixed with a scientific notation
        instead.

    """
    getlabel = getattr(ax, f"get_{axis}label")
    setlabel = getattr(ax, f"set_{axis}label")

    label = getlabel()
    unit = _unit_from_label(label)

    getattr(ax, f"{axis}axis").set_major_formatter(_SIFormatter(si))

    if prefix and (unit is not None):
        if power:
            setlabel(label.replace(f"({unit})", f"($\\times{{{10}}}^{{{si}}}$ {unit})"))
        else:
            setlabel(label.replace(f"({unit})", f"({get_si_str(si)}{unit})"))


def set_titles(axes, labels, order="C", **kwargs):
    axlist = np.array(axes, dtype=object).flatten(order=order)
    labels = np.asarray(labels)
    for ax, label in zip(axlist.flat, labels.flat):
        ax.set_title(label, **kwargs)


def set_xlabels(axes, labels, order="C", **kwargs):
    axlist = np.array(axes, dtype=object).flatten(order=order)
    if isinstance(labels, str):
        labels = [labels] * len(axlist)
    labels = np.asarray(labels)
    for ax, label in zip(axlist.flat, labels.flat):
        ax.set_xlabel(label, **kwargs)


def set_ylabels(axes, labels, order="C", **kwargs):
    axlist = np.array(axes, dtype=object).flatten(order=order)
    if isinstance(labels, str):
        labels = [labels] * len(axlist)
    labels = np.asarray(labels)
    for ax, label in zip(axlist.flat, labels.flat):
        ax.set_ylabel(label, **kwargs)


def sizebar(
    ax: matplotlib.axes.Axes,
    value: float,
    unit: str,
    si: int = 0,
    resolution: float = 1.0,
    decimals: int = 0,
    label: str | None = None,
    loc: Literal[
        "upper left",
        "upper center",
        "upper right",
        "center left",
        "center",
        "center right",
        "lower left",
        "lower center",
        "lower right",
    ] = "lower right",
    pad: float = 0.1,
    borderpad: float = 0.5,
    sep: float = 3.0,
    frameon: bool = False,
    **kwargs,
):
    """

    Parameters
    ----------
    ax
        The `matplotlib.axes.Axes` instance to place the size bar in.
    value
        Length of the size bar in terms of `unit`.
    unit
        An SI unit string without prefixes.
    si
        Exponents that have a corresponding SI prefix
    resolution
        Value to scale the data coordinates in terms of `unit`.
    decimals
        Number of decimals on the size bar label.
    label
        When provided, overrides the automatically generated label string.
    loc
        Location of the size bar.
    pad
        Padding around the label and size bar, in fraction of the font size.
    borderpad
        Border padding, in fraction of the font size.
    sep
        Separation between the label and the size bar, in points.
    frameon
        If True, draw a box around the horizontal bar and label.
    **kwargs
        Keyword arguments forwarded to
        `mpl_toolkits.axes_grid1.anchored_artists.AnchoredSizeBar`.

    """

    size = value / resolution
    unit = get_si_str(si) + unit
    value = np.around(value / 10**si, decimals=decimals)
    if int(value) == value:
        value = int(value)
    if label is None:
        label = f"{value} {unit}"

    kwargs.setdefault("color", axes_textcolor(ax))

    asb = AnchoredSizeBar(
        ax.transData,
        size,
        label,
        loc=loc,
        pad=pad,
        borderpad=borderpad,
        sep=sep,
        frameon=frameon,
        **kwargs,
    )
    ax.add_artist(asb)
    return asb



"""
Tools related to the real and reciprocal lattice.

.. currentmodule:: erlab.lattice

"""

__all__ = ["angle_between", "abc2avec", "avec2abc", "to_reciprocal", "to_real"]


import numpy as np
import numpy.typing as npt


def angle_between(v1: npt.NDArray[np.float64], v2: npt.NDArray[np.float64]) -> float:
    """Returns the angle between two vectors.

    Parameters
    ----------
    v1, v2:
        1D array of length 3, specifying a vector.

    Returns
    -------
    float
        The angle in degrees.
    """
    return np.rad2deg(np.arccos(np.clip(np.dot(v1, v2), -1.0, 1.0)))


def abc2avec(
    a: float, b: float, c: float, alpha: float, beta: float, gamma: float
) -> npt.NDArray:
    """Constructs lattice vectors from lattice parameters."""
    alpha, beta, gamma = np.deg2rad(alpha), np.deg2rad(beta), np.deg2rad(gamma)
    sa, ca, sb, cb, cg = (
        np.sin(alpha),
        np.cos(alpha),
        np.sin(beta),
        np.cos(beta),
        np.cos(gamma),
    )

    gp = np.arccos(np.clip((ca * cb - cg) / (sa * sb), -1.0, 1.0))
    cgp, sgp = np.cos(gp), np.sin(gp)
    return np.array(
        [
            [a * sb, 0, a * cb],
            [-b * sa * cgp, b * sa * sgp, b * ca],
            [0, 0, c],
        ]
    )


def avec2abc(
    avec: npt.NDArray[np.float64],
) -> tuple[float, float, float, float, float, float]:
    """Determines lattice parameters from lattice vectors."""
    a, b, c = (np.linalg.norm(x) for x in avec)
    alpha = angle_between(avec[1] / b, avec[2] / c)
    beta = angle_between(avec[2] / c, avec[0] / a)
    gamma = angle_between(avec[0] / a, avec[1] / b)
    return a, b, c, alpha, beta, gamma


def to_reciprocal(avec: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """Constructs the reciprocal lattice vectors from real lattice vectors.

    Parameters
    ----------
    avec
        Real lattice vectors, given as a 3 by 3 numpy array with each basis vector given
        in each row.

    Returns
    -------
    bvec
        The reciprocal lattice vectors.
    """
    return 2 * np.pi * np.linalg.inv(avec).T


def to_real(bvec: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """Constructs the real lattice vectors from reciprocal lattice vectors.

    Parameters
    ----------
    bvec
        Reciprocal lattice vectors, given as a 3 by 3 numpy array with each basis vector
        given in each row.

    Returns
    -------
    avec
        The real lattice vectors.
    """
    return np.linalg.inv(bvec.T / 2 / np.pi)



"""Momentum conversion functions.

Typically, the user will not have to call this module directly, but will instead use the
:func:`erlab.accessors.MomentumAccessor.convert` method.

Angle conventions and function forms are based on Ref. :cite:p:`Ishida2018`.

"""

import enum
from collections.abc import Callable

import numpy as np
import numpy.typing as npt

import erlab.constants
import erlab.io
import erlab.lattice


class AxesConfiguration(enum.IntEnum):
    """
    Enum class representing different types of axes configurations. See Ref.
    :cite:p:`Ishida2018`.

    """

    Type1 = 1
    Type2 = 2
    Type1DA = 3
    Type2DA = 4


def _sinc(x):
    return np.sin(x) / (x + 1e-15)


def kz_func_inv(kz, inner_potential, kx, ky):
    k_perp_sq = kx**2 + ky**2
    k_z_sq = kz**2
    return k_perp_sq + k_z_sq - inner_potential / erlab.constants.rel_kzconv


def kz_func(kinetic_energy, inner_potential, kx, ky):
    k_tot = erlab.constants.rel_kconv * np.sqrt(kinetic_energy)
    k_perp_sq = k_tot**2 - kx**2 - ky**2
    k_z_sq = k_perp_sq + inner_potential / erlab.constants.rel_kzconv
    return np.sqrt(np.clip(k_z_sq, a_min=0, a_max=None))


def get_kconv_func(
    kinetic_energy: np.floating | npt.NDArray,
    configuration: AxesConfiguration,
    angle_params: dict[str, float],
) -> tuple[Callable, Callable]:
    """
    Returns the appropriate momentum conversion functions based on the given
    configuration and kinetic energy.

    Parameters
    ----------
    kinetic_energy
        The kinetic energy in eV.
    configuration
        Experimental configuration.
    angle_params
        Dictionary of required angle parameters. If the configuration has a DA, the
        parameters should be `delta`, `chi`, `chi0`, `xi`, and `xi0`. Otherwise, they
        should be `delta`, `xi`, `xi0`, and `beta0`, following the notation in Ref.
        :cite:p:`Ishida2018`.

    Returns
    -------
    forward_func : Callable
        Forward function that takes :math:`(α, β)` and returns :math:`(k_x, k_y)`.
    inverse_func : Callable
        Inverse function that takes :math:`(k_x, k_y)` or :math:`(k_x, k_y, k_z)` and
        returns :math:`(α, β)`. If :math:`k_z` is given, it will return the angles
        broadcasted to :math:`k_z` instead of the provided kinetic energy.

    Raises
    ------
    ValueError
        If the given configuration is not valid.

    Note
    ----
    - The only requirement for the input parameters of the returned functions is that
      the shape of the input angles must be broadcastable with each other, and with the
      shape of the kinetic energy array. This means that the shape of the output array
      can be controlled By adjusting the shape of the input arrays. For instance, if the
      kinetic energy is given as a (L, 1, 1) array, :math:`k_x` as a (1, M, 1) array,
      and :math:`k_y` as a (1, 1, N) array, the output angle arrays :math:`α` and
      :math:`β` will each be broadcasted to a (L, M, N) array which can be directly used
      for interpolation.
    - However, the user will not have to worry about the shape of the input arrays,
      because using `xarray.DataArray` objects as the input will most likely broadcast
      the arrays automatically!

    See Also
    --------
    `NumPy Broadcasting Documentation
    <https://numpy.org/doc/stable/user/basics.broadcasting.html>`_

    """
    k_tot = erlab.constants.rel_kconv * np.sqrt(kinetic_energy)

    match configuration:
        case AxesConfiguration.Type1:
            func = _kconv_func_type1
        case AxesConfiguration.Type2:
            func = _kconv_func_type2
        case AxesConfiguration.Type1DA:
            func = _kconv_func_type1_da
        case AxesConfiguration.Type2DA:
            func = _kconv_func_type2_da
        case _:
            ValueError(f"Invalid configuration {configuration}")

    return func(k_tot, **angle_params)


def _kperp_func(k_tot_sq, kx, ky):
    r""":math:`\sqrt{k^2 - k_x^2 - k_y^2}`"""
    return np.sqrt(np.clip(k_tot_sq - kx**2 - ky**2, a_min=0, a_max=None))


def _kconv_func_type1(
    k_tot: np.floating | npt.NDArray,
    delta: np.floating = 0.0,
    xi: np.floating = 0.0,
    xi0: np.floating = 0.0,
    beta0: np.floating = 0.0,
):
    cd, sd = np.cos(np.deg2rad(delta)), np.sin(np.deg2rad(delta))  # δ
    cx, sx = np.cos(np.deg2rad(xi - xi0)), np.sin(np.deg2rad(xi - xi0))  # ξ - ξ0

    k_tot_sq = k_tot**2

    def _forward_func(alpha, beta):
        alpha_r, beta_r = np.deg2rad(alpha), np.deg2rad(beta - beta0)

        ca, cb = np.cos(alpha_r), np.cos(beta_r)
        sa, sb = np.sin(alpha_r), np.sin(beta_r)

        kx = k_tot * ((sd * sb + cd * sx * cb) * ca - cd * cx * sa)
        ky = k_tot * ((-cd * sb + sd * sx * cb) * ca - sd * cx * sa)

        return kx, ky

    def _inverse_func(kx, ky, kz=None):
        # cx, sx = np.cos(np.deg2rad(xi)), np.sin(np.deg2rad(xi))  # ξ
        # mask = kx**2 + ky**2 > k_tot_sq + 1e-15
        # kx, ky = np.ma.masked_where(mask, kx), np.ma.masked_where(mask, ky)
        if kz is None:
            k_sq = k_tot_sq
            k = k_tot
        else:
            k_sq = kx**2 + ky**2 + kz**2
            k = np.sqrt(k_sq)

        kperp = _kperp_func(k_sq, kx, ky)

        alpha = np.arcsin((sx * kperp - cx * (cd * kx + sd * ky)) / k)
        beta = np.arctan((sd * kx - cd * ky) / (sx * (cd * kx + sd * ky) + cx * kperp))

        return np.rad2deg(alpha), np.rad2deg(beta) + beta0

    return _forward_func, _inverse_func


def _kconv_func_type2(
    k_tot: np.floating | npt.NDArray,
    delta: np.floating = 0.0,
    xi: np.floating = 0.0,
    xi0: np.floating = 0.0,
    beta0: np.floating = 0.0,
):
    cd, sd = np.cos(np.deg2rad(delta)), np.sin(np.deg2rad(delta))  # δ
    cx, sx = np.cos(np.deg2rad(xi - xi0)), np.sin(np.deg2rad(xi - xi0))  # ξ - ξ0

    k_tot_sq = k_tot**2

    def _forward_func(alpha, beta):
        alpha_r, beta_r = np.deg2rad(alpha), np.deg2rad(beta - beta0)

        ca, sa, sb = np.cos(alpha_r), np.sin(alpha_r), np.sin(beta_r)

        kx = k_tot * ((sd * sx + cd * sb * cx) * ca - (sd * cx - cd * sb * sx) * sa)
        ky = k_tot * ((-cd * sx + sd * sb * cx) * ca + (cd * cx + sd * sb * sx) * sa)

        return kx, ky

    def _inverse_func(kx, ky, kz=None):
        # cx, sx = np.cos(np.deg2rad(xi)), np.sin(np.deg2rad(xi))  # ξ
        # mask = kx**2 + ky**2 > k_tot_sq + 1e-15
        # kx, ky = np.ma.masked_where(mask, kx), np.ma.masked_where(mask, ky)
        if kz is None:
            k_sq = k_tot_sq
            k = k_tot
        else:
            k_sq = kx**2 + ky**2 + kz**2
            k = np.sqrt(k_sq)

        kperp = _kperp_func(k_sq, kx, ky)
        kproj = np.sqrt(np.clip(k_sq - (sd * kx - cd * ky) ** 2, a_min=0, a_max=None))

        alpha = np.arcsin((sx * kproj - cx * (sd * kx - cd * ky)) / k)
        beta = np.arctan((cd * kx + sd * ky) / kperp)

        return np.rad2deg(alpha), np.rad2deg(beta) + beta0

    return _forward_func, _inverse_func


def _kconv_func_type1_da(
    k_tot: np.floating | npt.NDArray,
    delta: np.floating = 0.0,
    chi: np.floating = 0.0,
    chi0: np.floating = 0.0,
    xi: np.floating = 0.0,
    xi0: np.floating = 0.0,
):
    _fwd_2, _inv_2 = _kconv_func_type2_da(k_tot, delta, chi, chi0, xi, xi0)

    def _forward_func(alpha, beta):
        return _fwd_2(-beta, alpha)

    def _inverse_func(kx, ky, kz=None):
        alpha, beta = _inv_2(kx, ky, kz)
        return beta, -alpha

    return _forward_func, _inverse_func


def _kconv_func_type2_da(
    k_tot: np.floating | npt.NDArray,
    delta: np.floating = 0.0,
    chi: np.floating = 0.0,
    chi0: np.floating = 0.0,
    xi: np.floating = 0.0,
    xi0: np.floating = 0.0,
):
    cd, sd = np.cos(np.deg2rad(delta)), np.sin(np.deg2rad(delta))  # δ, azimuth
    cx, sx = np.cos(np.deg2rad(xi - xi0)), np.sin(np.deg2rad(xi - xi0))  # ξ
    cc, sc = np.cos(np.deg2rad(chi - chi0)), np.sin(np.deg2rad(chi - chi0))  # χ

    t11, t12, t13 = cx * cd, cx * sd, -sx
    t21, t22, t23 = sc * sx * cd - cc * sd, sc * sx * sd + cc * cd, sc * cx
    t31, t32, t33 = cc * sx * cd + sc * sd, cc * sx * sd - sc * cd, cc * cx

    k_tot_sq = k_tot**2

    def _forward_func(alpha, beta):
        ar, br = np.deg2rad(alpha), np.deg2rad(beta)

        absq = np.sqrt(ar**2 + br**2)
        scab, cab = _sinc(absq), np.cos(absq)

        # Type I DA
        # kx = k_tot * (
        #     (-ar * cd * cx + br * sd * cc - br * cd * sx * sc) * scab
        #     + (sd * sc + cd * sx * cc) * cab
        # )
        # ky = k_tot * (
        #     (-ar * sd * cx - br * cd * cc - br * sd * sx * sc) * scab
        #     - (cd * sc - sd * sx * cc) * cab
        # )

        # Type II DA
        kx = k_tot * (
            (-br * cd * cx - ar * sd * cc + ar * cd * sx * sc) * scab
            + (sd * sc + cd * sx * cc) * cab
        )
        ky = k_tot * (
            (-br * sd * cx + ar * cd * cc + ar * sd * sx * sc) * scab
            - (cd * sc - sd * sx * cc) * cab
        )

        return kx, ky

    def _inverse_func(kx, ky, kz=None):
        # mask = kx**2 + ky**2 > k_tot_sq + 1e-15
        # kx, ky = np.ma.masked_where(mask, kx), np.ma.masked_where(mask, ky)

        if kz is None:
            k_sq = k_tot_sq
            k = k_tot
        else:
            k_sq = kx**2 + ky**2 + kz**2
            k = np.sqrt(k_sq)

        kperp = _kperp_func(k_sq, kx, ky)  # sqrt(k² − k_x² − k_y²)

        proj1 = t11 * kx + t12 * ky + t13 * kperp
        proj2 = t21 * kx + t22 * ky + t23 * kperp
        proj3 = t31 * kx + t32 * ky + t33 * kperp

        # Type I DA
        # alpha = (
        #     -np.arccos(np.clip(proj3 / k_tot, -1, 1))
        #     * proj1
        #     / np.sqrt((k_sq - proj3**2).clip(min=0))
        # )
        # beta = (
        #     -np.arccos(np.clip(proj3 / k_tot, -1, 1))
        #     * proj2
        #     / np.sqrt((k_sq - proj3**2).clip(min=0))
        # )

        # Type II DA
        alpha = (
            np.arccos(np.clip(proj3 / k, -1, 1))
            * proj2
            / np.sqrt((k_sq - proj3**2).clip(min=0))
        )
        beta = (
            -np.arccos(np.clip(proj3 / k, -1, 1))
            * proj1
            / np.sqrt((k_sq - proj3**2).clip(min=0))
        )

        return np.rad2deg(alpha), np.rad2deg(beta)

    return _forward_func, _inverse_func



"""
Helper functions for parallel processing.

.. currentmodule:: erlab.parallel

"""

__all__ = ["joblib_progress", "joblib_progress_qt"]

import contextlib
import sys

import joblib
import joblib._parallel_backends
import tqdm
import tqdm.notebook
from qtpy import QtCore


def is_notebook():
    # http://stackoverflow.com/questions/34091701/determine-if-were-in-an-ipython-notebook-session
    if "IPython" not in sys.modules:  # IPython hasn't been imported
        return False
    from IPython import get_ipython

    # check for `kernel` attribute on the IPython instance
    return getattr(get_ipython(), "kernel", None) is not None


@contextlib.contextmanager
def joblib_progress(file=None, notebook=None, dynamic_ncols=True, **kwargs):
    """Context manager to patch joblib to report into tqdm progress bar given as
    argument"""

    if file is None:
        file = sys.stdout

    if notebook is None:
        notebook = is_notebook()

    if notebook:
        tqdm_object = tqdm.notebook.tqdm(
            iterable=None, dynamic_ncols=dynamic_ncols, file=file, **kwargs
        )
    else:
        tqdm_object = tqdm.tqdm(
            iterable=None, dynamic_ncols=dynamic_ncols, file=file, **kwargs
        )

    def tqdm_print_progress(self):
        if self.n_completed_tasks > tqdm_object.n:
            n_completed = self.n_completed_tasks - tqdm_object.n
            tqdm_object.update(n=n_completed)

    original_print_progress = joblib.parallel.Parallel.print_progress
    joblib.parallel.Parallel.print_progress = tqdm_print_progress

    try:
        yield tqdm_object
    finally:
        joblib.parallel.Parallel.print_progress = original_print_progress
        tqdm_object.close()


@contextlib.contextmanager
def joblib_progress_qt(signal: QtCore.Signal):
    """Context manager for interactive windows.

    The number of completed tasks are emitted by the given signal.
    """

    def qt_print_progress(self):
        signal.emit(self.n_completed_tasks)

    original_print_progress = joblib.parallel.Parallel.print_progress
    joblib.parallel.Parallel.print_progress = qt_print_progress

    try:
        yield None
    finally:
        joblib.parallel.Parallel.print_progress = original_print_progress



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
# -- Imports -----------------------------------------------------------------

import importlib.metadata
import os

import pybtex.plugin
import pybtex.style.formatting
import pybtex.style.formatting.unsrt
import pybtex.style.template

# Build docs with PyQt6 since PySide6 is broken
# https://bugreports.qt.io/browse/PYSIDE-1884

os.environ["QT_API"] = "pyqt6"


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information


project = "ERLab"
copyright = "2023, Kimoon Han"
author = "Kimoon Han"
release = importlib.metadata.version("erlab")
version = release


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    # "sphinx_autodoc_typehints",
    "sphinx.ext.viewcode",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "matplotlib.sphinxext.plot_directive",
    "matplotlib.sphinxext.figmpl_directive",
    # "IPython.sphinxext.ipython_directive",
    # "IPython.sphinxext.ipython_console_highlighting",
    # "sphinx.ext.inheritance_diagram",
    "sphinxcontrib.bibtex",
    "sphinx_qt_documentation",
    "sphinx_copybutton",
    "nbsphinx",
    "sphinx_design",
]


templates_path = ["_templates"]
exclude_patterns = []

default_role = "obj"

# nitpicky = False
# nitpick_ignore = [("py:class", "numpy.float64")]


# -- Autosummary and autodoc settings ----------------------------------------

autosummary_generate = True
autosummary_generate_overwrite = True

autodoc_inherit_docstrings = False
autodoc_typehints = "both"
autodoc_typehints_format = "short"
autodoc_class_signature = "mixed"
autodoc_member_order = "bysource"
autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    # "exclude-members":("sigDataChanged",),
    "show-inheritance": False,
}
autodoc_typehints_description_target = "all"

# autodoc_type_aliases = {
# "numpy.float64": "float",
# "float64": "float",
# }

# -- Napoleon settings -------------------------------------------------------

napoleon_google_docstring = False
napoleon_numpy_docstring = True
# napoleon_include_init_with_doc = False
# napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
# napoleon_use_admonition_for_examples = False
# napoleon_use_admonition_for_notes = False
# napoleon_use_admonition_for_references = False
# napoleon_use_ivar = False
napoleon_use_param = True
# napoleon_use_rtype = True
napoleon_preprocess_types = True
# napoleon_type_aliases = {
#     "array-like": "ndarray <numpy.ndarray>",
#     "numpy.float64": "float",
#     "float64": "float",
# }
napoleon_attr_annotations = True
napoleon_custom_sections = [("Signals", "params_style")]

# -- nbsphinx options --------------------------------------------------------

nbsphinx_execute_arguments = [
    "--InlineBackend.figure_formats={'svg', 'pdf'}",
]

# -- Qt documentation & intersphinx ------------------------------------------

qt_documentation = "PyQt6"
intersphinx_mapping = {
    "PyQt5": ("https://doc.qt.io/qtforpython-6/", None),
    "python": ("https://docs.python.org/3.10/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "numba": ("https://numba.readthedocs.io/en/stable/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "lmfit": ("https://lmfit.github.io/lmfit-py/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "xarray": ("https://docs.xarray.dev/en/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
    "pyqtgraph": ("https://pyqtgraph.readthedocs.io/en/latest/", None),
    "arpes": ("https://arpes.readthedocs.io/en/latest/", None),
    "csaps": ("https://csaps.readthedocs.io/en/latest/", None),
    "iminuit": ("https://scikit-hep.org/iminuit/", None),
}


# -- Plot configuration ------------------------------------------------------

plot_formats = ["pdf"]
plot_basedir = "pyplots"
plot_html_show_formats = False

# -- Misc. settings ----------------------------------------------------------

copybutton_exclude = ".linenos, .gp, .go"

# -- Bibliography settings ---------------------------------------------------

pybtex.style.formatting.unsrt.date = pybtex.style.template.words(sep="")[
    "(", pybtex.style.template.field("year"), ")"
]


class ApsStyle(pybtex.style.formatting.unsrt.Style):
    """
    APS style for BibTeX formatting, adapted from the conf.py file of the `mitiq
    library<https://github.com/unitaryfund/mitiq>`_.
    """

    def format_title(self, e, which_field, as_sentence=True):
        formatted_title = pybtex.style.template.field(
            which_field, apply_func=lambda text: text.capitalize()
        )
        formatted_title = pybtex.style.template.tag("em")[formatted_title]
        if as_sentence:
            return pybtex.style.template.sentence[formatted_title]
        else:
            return formatted_title

    def get_article_template(self, e):
        volume_and_pages = pybtex.style.template.first_of[
            # volume and pages
            pybtex.style.template.optional[
                pybtex.style.template.join[
                    " ",
                    pybtex.style.template.tag("strong")[
                        pybtex.style.template.field("volume")
                    ],
                    ", ",
                    pybtex.style.template.field(
                        "pages",
                        apply_func=pybtex.style.formatting.unsrt.dashify,
                    ),
                ],
            ],
            # pages only
            pybtex.style.template.words[
                "pages",
                pybtex.style.template.field(
                    "pages", apply_func=pybtex.style.formatting.unsrt.dashify
                ),
            ],
        ]
        template = pybtex.style.formatting.toplevel[
            self.format_names("author"),
            self.format_title(e, "title"),
            pybtex.style.template.sentence(sep=" ")[
                pybtex.style.template.field("journal"),
                pybtex.style.template.optional[volume_and_pages],
                pybtex.style.formatting.unsrt.date,
            ],
            self.format_web_refs(e),
        ]
        return template

    def get_book_template(self, e):
        template = pybtex.style.formatting.toplevel[
            self.format_author_or_editor(e),
            self.format_btitle(e, "title"),
            self.format_volume_and_series(e),
            pybtex.style.template.sentence(sep=" ")[
                pybtex.style.template.sentence(add_period=False)[
                    pybtex.style.template.field("publisher"),
                    pybtex.style.template.optional_field("address"),
                    self.format_edition(e),
                ],
                pybtex.style.formatting.unsrt.date,
            ],
            pybtex.style.template.optional[
                pybtex.style.template.sentence[self.format_isbn(e)]
            ],
            pybtex.style.template.sentence[
                pybtex.style.template.optional_field("note")
            ],
            self.format_web_refs(e),
        ]
        return template

    def get_incollection_template(self, e):
        template = pybtex.style.formatting.toplevel[
            pybtex.style.template.sentence[self.format_names("author")],
            self.format_title(e, "title"),
            pybtex.style.template.words[
                "In",
                pybtex.style.template.sentence[
                    pybtex.style.template.optional[
                        self.format_editor(e, as_sentence=False)
                    ],
                    self.format_btitle(e, "booktitle", as_sentence=False),
                    self.format_volume_and_series(e, as_sentence=False),
                    self.format_chapter_and_pages(e),
                ],
            ],
            pybtex.style.template.sentence(sep=" ")[
                pybtex.style.template.sentence(add_period=False)[
                    pybtex.style.template.optional_field("publisher"),
                    pybtex.style.template.optional_field("address"),
                    self.format_edition(e),
                ],
                pybtex.style.formatting.unsrt.date,
            ],
            self.format_web_refs(e),
        ]
        return template


pybtex.plugin.register_plugin("pybtex.style.formatting", "apsstyle", ApsStyle)

bibtex_bibfiles = ["refs.bib"]
bibtex_default_style = "apsstyle"
bibtex_footbibliography_header = ".. rubric:: References"


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
pygments_dark_style = "monokai"
language = "en"

html_static_path = ["_static"]
html_css_files = ["pied-piper-admonition.css"]
html_theme_options: dict[str, object] = {
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/kmnhan/erlabpy",
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,
            "class": "",
        },
    ],
    "light_css_variables": {
        "color-brand-primary": "#6d50bf",
        "color-brand-content": "#6d50bf",
    },
    "dark_css_variables": {
        "color-brand-primary": "#a180ff",
        "color-brand-content": "#a180ff",
    },
}

# -- LaTeX options -----------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#latex-options

latex_engine = "lualatex"
latex_table_style = ["booktabs"]
latex_elements = {
    "fontpkg": r"""\usepackage{fontspec,unicode-math}
""",
}



"""
Classes and functions for plotting atoms and bonds in a crystal structure using
matplotlib's 3D plotting capabilities.

Some of the projection code was adapted from kwant.

"""

import contextlib
import functools
import itertools
from collections.abc import Callable, Sequence
from typing import Literal

import matplotlib.collections
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d
import mpl_toolkits.mplot3d.art3d
import mpl_toolkits.mplot3d.proj3d
import numpy as np
import numpy.typing as npt

__all__ = ["Atom3DCollection", "Bond3DCollection", "CrystalProperty"]


def sunflower_sphere(n: int = 100) -> npt.NDArray[np.float64]:
    """Returns n points on a sphere using the sunflower algorithm."""
    indices = np.arange(0, n, dtype=float) + 0.5

    phi = np.arccos(1 - 2 * indices / n)
    theta = np.pi * (1 + 5**0.5) * indices

    return np.c_[
        np.cos(theta) * np.sin(phi), np.sin(theta) * np.sin(phi), np.cos(phi)
    ].T


unit_sphere: npt.NDArray[np.float64] = sunflower_sphere(100)


def projected_length(
    ax: mpl_toolkits.mplot3d.Axes3D, length: np.float64 | Sequence[np.float64]
):
    if np.iterable(length):
        return np.asarray([projected_length(ax, d) for d in np.asarray(length).flat])

    # rc = np.array([ax.get_xlim3d(), ax.get_ylim3d(), ax.get_zlim3d()]).sum(1) / 2

    rc = np.array(
        mpl_toolkits.mplot3d.proj3d.inv_transform(
            0.5,
            0.5,
            0.5,
            mpl_toolkits.mplot3d.proj3d.world_transformation(
                *ax.get_xlim3d(),
                *ax.get_ylim3d(),
                *ax.get_zlim3d(),
                pb_aspect=ax._roll_to_vertical(ax._box_aspect),
            ),
        )
    )
    rs = unit_sphere * length + rc.reshape(-1, 1)

    transform = mpl_toolkits.mplot3d.proj3d.proj_transform
    proj = ax.get_proj().round(15)
    rp = np.asarray(transform(*rs, proj)[:2])
    rc[:2] = transform(*rc, proj)[:2]

    coords = rp - np.repeat(rc[:2].reshape(-1, 1), len(rs[0]), axis=1)
    return np.linalg.norm(coords, axis=0).mean()


def projected_length_pos(ax: mpl_toolkits.mplot3d.Axes3D, length, position):
    if position.ndim > 1:
        if not np.iterable(length):
            length = np.ones(len(position)) * length
        return np.asarray(
            [
                projected_length_pos(ax, d, p)
                for d, p in zip(np.asarray(length).flat, np.asarray(position))
            ]
        )
    rc = np.asarray(position).reshape(-1, 1)
    rs = unit_sphere * length + rc

    transform = mpl_toolkits.mplot3d.proj3d.proj_transform
    proj = ax.get_proj()
    rp = np.asarray(transform(*rs, proj)[:2])
    rc[:2] = transform(*rc, proj)[:2]

    coords = rp - rc[:2]
    return np.linalg.norm(coords, axis=0).mean()


def _zalpha(colors, zs):
    """Modified from `mpl_toolkits.mplot3d.art3d` to modifies the brightness of the
    color based on depth, rather than setting the transparency."""
    if len(colors) == 0 or len(zs) == 0:
        return np.zeros((0, 4))
    norm = plt.Normalize(min(zs), max(zs))
    sats = 1 - norm(zs) * 0.7
    rgba = np.broadcast_to(mcolors.to_rgba_array(colors), (len(zs), 4))
    rgba = rgba.T * sats
    rgba += 1 - sats
    rgba = rgba.T
    rgba[:, 3] = 1
    return rgba


class Atom3DCollection(mpl_toolkits.mplot3d.art3d.Path3DCollection):
    """A collection of 3D scatter points that represents atoms in a crystal structure.

    This class is not for direct instantiation, but is used to patch the collection
    returned by the `scatter` method. See the implementation of `CrystalProperty.plot`
    for usage.

    """

    def __init__(self, *args, scale_size: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        self._scale_size = scale_size

    def _maybe_depth_shade_and_sort_colors(self, color_array):
        if self._alpha == 1.0:
            return super()._maybe_depth_shade_and_sort_colors(color_array)
        color_array = (
            _zalpha(color_array, self._vzs)
            if self._vzs is not None and self._depthshade
            else color_array
        )
        if len(color_array) > 1:
            color_array = color_array[self._z_markers_idx]
        return mcolors.to_rgba_array(color_array, self._alpha)

    def set_sizes(self, sizes: np.ndarray, dpi: float = 72.0):
        super().set_sizes(sizes, dpi)
        self.sizes_orig = np.asarray(sizes) if np.iterable(sizes) else sizes

    @contextlib.contextmanager
    def _use_zordered_offset(self):
        if self._offset_zordered is None:
            yield
        else:
            old_offset = self._offsets
            super().set_offsets(self._offset_zordered)
            old_sizes = self._sizes
            super().set_sizes(
                self._sizes[
                    old_offset[:, 0].argsort()[
                        self._offset_zordered[:, 0].argsort().argsort()
                    ]
                ],
                self.figure.dpi,
            )
            try:
                yield
            finally:
                self._sizes = old_sizes
                self._offsets = old_offset

    def draw(self, renderer):
        if self._scale_size:  # and np.isfinite(self.axes._focal_length):
            # proj_sizes = projected_length(self.axes, np.sqrt(self.sizes_orig))
            proj_sizes = projected_length_pos(
                self.axes, np.sqrt(self.sizes_orig), np.c_[self._offsets3d]
            )
            args = self.axes.transData.frozen().to_values()
            proj_sizes = proj_sizes * ((args[0] + args[3]) / 2) * 72.0 / self.figure.dpi
        else:
            proj_sizes = np.sqrt(self.sizes_orig)

        super().set_sizes(proj_sizes**2, self.figure.dpi)
        with self._use_zordered_offset():
            with matplotlib.cbook._setattr_cm(self, _in_draw=True):
                matplotlib.collections.Collection.draw(self, renderer)

    # def draw(self, renderer):
    #     # Note: unlike in the 2D case, where we can enforce equal
    #     #       aspect ratio, this (currently) does not work with
    #     #       3D plots in matplotlib. As an approximation, we
    #     #       thus scale with the average of the x- and y-axis
    #     #       transformation.
    #     self._factor = (
    #         self._sz_real * ((args[0] + args[3]) / 2) * 72.0 / self.figure.dpi
    #     )
    #     super().draw(renderer)


class Bond3DCollection(mpl_toolkits.mplot3d.art3d.Line3DCollection):
    """
    A subclass of `mpl_toolkits.mplot3d.art3d.Line3DCollection` representing a
    collection of 3D bonds.

    Parameters
    ----------
    segments
        List of segments representing the bonds.
    scale_linewidths
        Boolean indicating whether to scale the linewidths based on the plot's
        perspective.
    **kwargs
        Additional keyword arguments to be passed to
        `mpl_toolkits.mplot3d.art3d.Line3DCollection`.

    """

    def __init__(self, segments, *, scale_linewidths: bool = True, **kwargs: dict):
        super().__init__(segments, **kwargs)
        self._scale_linewidths: bool = scale_linewidths

    def set_linewidth(self, lw):
        super().set_linewidth(lw)
        self.linewidths_orig = np.asarray(lw) if np.iterable(lw) else lw

    def draw(self, renderer):
        if self._scale_linewidths:  # and np.isfinite(self.axes._focal_length):
            proj_len = projected_length(self.axes, self.linewidths_orig)
            args = self.axes.transData.frozen().to_values()
            proj_len = proj_len * ((args[0] + args[3]) / 2) * 72.0 / self.figure.dpi
        else:
            proj_len = self.linewidths_orig

        super().set_linewidth(proj_len)
        super().draw(renderer)

        # if sizes is None:
        #     self._sizes = np.array([])
        #     self._transforms = np.empty((0, 3, 3))
        # else:
        #     self._sizes = np.asarray(sizes)
        #     self._transforms = np.zeros((len(self._sizes), 3, 3))
        #     scale = np.sqrt(self._sizes) * dpi / 72.0 * self._factor
        #     self._transforms[:, 0, 0] = scale
        #     self._transforms[:, 1, 1] = scale
        #     self._transforms[:, 2, 2] = 1.0
        # self.stale = True


class CrystalProperty:
    def __init__(
        self,
        atom_pos: dict[
            str, npt.NDArray[np.float64] | Sequence[npt.NDArray[np.float64]]
        ],
        avec: npt.NDArray[np.float64],
        offset: npt.NDArray[np.float64] | Sequence[float] = (0.0, 0.0, 0.0),
        radii: Sequence[float] | None = None,
        colors: Sequence[str | tuple[float, ...]] | None = None,
        repeat: tuple[int, int, int] = (1, 1, 1),
        bounds: dict[Literal["x", "y", "z"], tuple[float, float]] | None = None,
        mask: Callable | None = None,
        r_factor: float = 0.4,
    ):
        """
        A class that stores the information required to plot a three dimensional crystal
        structure. The crystal can be repeated along the lattice vectors, and can be
        masked. Bonds can be added between atoms.

        Parameters
        ----------
        atom_pos
            A dictionary of atom positions in the unit cell, where the keys are the atom
            names and the values are the positions of the atoms. The positions can be
            given as a (N, 3) numpy array or a list of N numpy arrays of shape (3,).
        avec
            Lattice vectors of the crystal.
        offset
            Relative offset of the crystal in the plot, by default (0.0, 0.0, 0.0)
        radii
            Sequence of radii for each atom, must be the same length as the number of
            keys given in `atom_pos`. If not provided, all atoms will be given the same
            radius.
        colors
            Sequence of valid matplotlib colors for each atom, must be the same length
            as the number of keys given in `atom_pos`. If not provided, atoms will be
            colored with the current matplotlib prop_cycle in the order they are given
            in `atom_pos`.
        repeat
            Number of times to repeat the crystal in each direction, by default (1, 1,
            1)
        bounds
            A dictionary of bounds for the crystal in each direction. Atoms that exceed
            the bounds will not be plotted
        mask
            A function that takes three cartesian coordinates as the input and returns a
            boolean. If the function returns False, the atom will not be plotted, by
            default `None`
        r_factor
            Additional scaling factor for the bond lengths, by default 0.4

        """
        self.atom_pos_given = atom_pos
        self.avec = avec
        self.offset = np.asarray(offset)

        if radii is None:
            radii = [1.0] * len(self.atoms)
        self.atom_radii: dict[str, float] = dict(zip(self.atoms, radii))

        if colors is None:
            colors = [f"C{i}" for i in range(len(self.atoms))]
        self.atom_color: dict[str, str] = {
            k: mcolors.to_hex(v) for k, v in zip(self.atoms, colors)
        }

        self.repeat: tuple[int, int, int] = repeat
        self._bounds: dict[Literal["x", "y", "z"], tuple[float, float]] = (
            {} if bounds is None else bounds
        )
        self.mask: Callable | None = mask

        self.segments: list[
            tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]
        ] = []
        self._bond_lw: list[float] = []
        self._bond_c: list[str | tuple[float, ...]] = []

        self.r_factor: float = r_factor

    @classmethod
    def from_fractional(
        cls,
        frac_pos: dict[str, list[tuple[float, float, float]]],
        avec: npt.NDArray[np.float64],
        *args,
        **kwargs,
    ):
        atom_pos = {}
        for k, v in frac_pos.items():
            atom_pos[k] = [x[0] * avec[0] + x[1] * avec[1] + x[2] * avec[2] for x in v]
        return cls(atom_pos, avec, *args, **kwargs)

    @property
    def bounds(self) -> list[tuple[float, float]]:
        bound_list = []
        for dim in ("x", "y", "z"):
            try:
                bound_list.append(self._bounds[dim])
            except KeyError:
                bound_list.append((-np.inf, np.inf))
        return bound_list

    @property
    def atoms(self) -> list[str]:
        return list(self.atom_pos_given.keys())

    @functools.cached_property
    def atom_pos(self) -> dict[str, npt.NDArray[np.float64]]:
        atom_pos = {k: np.zeros((1, 3)) for k in self.atom_pos_given}
        for k, v in self.atom_pos_given.items():
            for x, y, z in itertools.product(*[range(-n + 1, n) for n in self.repeat]):
                atom_pos[k] = np.r_[
                    atom_pos[k],
                    v
                    + x * self.avec[0]
                    + y * self.avec[1]
                    + z * self.avec[2]
                    + self.offset,
                ]
            atom_pos[k] = atom_pos[k][1:]

        # Remove duplicates
        for k, v in atom_pos.items():
            atom_pos[k] = v[np.unique(v.round(5), axis=0, return_index=True)[1]]

        xlim, ylim, zlim = self.bounds
        valid = {}
        for k, v in atom_pos.items():
            mask_arr = (
                (xlim[0] < v[:, 0])
                * (v[:, 0] < xlim[1])
                * (ylim[0] < v[:, 1])
                * (v[:, 1] < ylim[1])
                * (zlim[0] < v[:, 2])
                * (v[:, 2] < zlim[1])
            )
            if self.mask is not None:
                mask_arr = mask_arr * self.mask(v[:, 0], v[:, 1], v[:, 2])
            valid[k] = np.where(mask_arr)
        masked_atom_pos = {}
        for k, v in atom_pos.items():
            masked_atom_pos[k] = np.asarray(
                [p for i, p in enumerate(v) if i in valid[k][0]]
            )
        return masked_atom_pos

    @property
    def _color_array(self) -> tuple[npt.NDArray[np.str_]]:
        clr_arr = []
        for k, v in self.atom_pos.items():
            for _ in range(len(v)):
                clr_arr.append(self.atom_color[k])
        return np.asarray(clr_arr)

    @property
    def _size_array(self) -> tuple[npt.NDArray[np.float64]]:
        sz_arr = []
        for k, v in self.atom_pos.items():
            for _ in range(len(v)):
                sz_arr.append(self.atom_radii[k])
        return np.asarray(sz_arr)

    @property
    def _atom_pos_array(self) -> npt.NDArray[np.float64]:
        return np.r_[*list(self.atom_pos.values())]

    def clear_bonds(self) -> None:
        self.segments.clear()
        self._bond_lw.clear()
        self._bond_c.clear()

    def add_bonds(
        self,
        atom1: str,
        atom2: str,
        min_length: float = 0.0,
        max_length: float = 2.6,
        linewidth: float = 0.25,
        color: str | tuple[float, ...] | None = None,
    ) -> None:
        new_seg = self._generate_segments(atom1, atom2, min_length, max_length)
        self.segments += new_seg
        self._bond_lw += [linewidth] * len(new_seg)
        self._bond_c += ["#b2b2b2" if color is None else color] * len(new_seg)

    def plot(
        self,
        ax: mpl_toolkits.mplot3d.Axes3D | None = None,
        scale_bonds: bool = True,
        scale_atoms: bool = True,
        clean_axes: bool = True,
        bond_kw: dict | None = None,
        atom_kw: dict | None = None,
    ):
        """
        Plot the crystal structure.

        Parameters
        ----------
        ax
            A 3D axes object to plot the crystal on. If not provided, add_subplot will
            be called on the current figure.
        scale_bonds
            Whether to scale the bond linewidths based on the distance from the camera,
            by default True
        scale_atoms
            Whether to scale the atom sizes based on the distance from the camera, by
            default True
        clean_axes
            Whether to clean the axes by removing the background and grid, setting pane
            color, and removing the margins, by default True
        bond_kw
            Keyword arguments passed onto `Bond3DCollection`
        atom_kw
            Keyword arguments passed onto `mpl_toolkits.mplot3d.Axes3D.scatter` used to
            plot the atoms.

        """

        if ax is None:
            ax = plt.gcf().add_subplot(projection="3d")

        if clean_axes:
            ax.set_facecolor("none")
            ax.xaxis.set_pane_color((1, 1, 1, 0))
            ax.yaxis.set_pane_color((1, 1, 1, 0))
            ax.zaxis.set_pane_color((1, 1, 1, 0))
            ax.grid(False)
            ax.set_xmargin(0)
            ax.set_ymargin(0)

        if bond_kw is None:
            bond_kw = {}
        if atom_kw is None:
            atom_kw = {}

        bond_kw.setdefault("linewidths", self._bond_lw)
        bond_kw.setdefault("colors", self._bond_c)
        bond_kw.setdefault("clip_on", False)
        bond_collection = Bond3DCollection(
            self.segments, scale_linewidths=scale_bonds, **bond_kw
        )
        ax.add_collection(bond_collection)

        atom_kw.setdefault("c", self._color_array)
        atom_kw.setdefault("edgecolors", "none")
        atom_kw.setdefault("clip_on", False)
        atom_kw.setdefault("lw", 0)

        ax.figure.canvas.draw()
        sc = ax.scatter(*self._atom_pos_array.T, **atom_kw)
        sc.__class__ = Atom3DCollection
        sc._scale_size = scale_atoms
        sc.set_sizes(self._size_array**2)

    def _generate_segments(
        self, atom1: str, atom2: str, min_length: float, max_length: float
    ) -> list[tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]]:
        atom1_pos, atom2_pos = self.atom_pos[atom1], self.atom_pos[atom2]
        r1, r2 = self.atom_radii[atom1], self.atom_radii[atom2]
        segments = []
        for p1 in atom1_pos:
            for p2 in atom2_pos:
                vec = p2 - p1
                d = np.linalg.norm(vec)
                vec_u = vec / d
                if d >= min_length and d <= max_length:
                    segments.append(
                        (
                            p1 + vec_u * r1 * self.r_factor,
                            p2 - vec_u * r2 * self.r_factor,
                        )
                    )
        return segments



"""
Everything related to plotting.

.. currentmodule:: erlab.plotting

Modules
=======

.. autosummary::
   :toctree: generated

   annotations
   atoms
   bz
   colors
   erplot
   general
   plot3d

"""

__all__ = [
    "annotations",
    "atoms",
    "bz",
    "colors",
    "general",
    "plot3d",
]

import importlib
import io
import os
import pkgutil

import matplotlib
import matplotlib.colors
import matplotlib.font_manager
import matplotlib.style
import numpy as np

import erlab.io

# Import colormaps if available
if importlib.util.find_spec("cmasher"):
    importlib.import_module("cmasher")
if importlib.util.find_spec("cmocean"):
    importlib.import_module("cmocean")
if importlib.util.find_spec("colorcet"):
    importlib.import_module("colorcet")


def load_igor_ct(fname, name):
    file = pkgutil.get_data(__package__, "IgorCT/" + fname)
    if fname.endswith(".txt"):
        values = np.genfromtxt(io.StringIO(file.decode()))
    elif fname.endswith(".ibw"):
        values = erlab.io.load_wave(io.BytesIO(file)).values

    cmap = matplotlib.colors.LinearSegmentedColormap.from_list(name, values / 65535)
    matplotlib.colormaps.register(cmap)
    matplotlib.colormaps.register(cmap.reversed())


load_igor_ct("CTBlueWhite.ibw", "BuWh")
load_igor_ct("CTRainbowLIght.ibw", "RainbowLight")
# load_igor_ct("CTRedTemperature.ibw", "RedTemperature")
load_igor_ct("ColdWarm.ibw", "ColdWarm")
load_igor_ct("BlueHot.ibw", "BlueHot")
# load_igor_ct("PlanetEarth.ibw", "PlanetEarth")
# load_igor_ct("ametrine.ibw", "ametrine")
# load_igor_ct("isolum.ibw", "isolum")
# load_igor_ct("morgenstemning.ibw", "morgenstemning")


matplotlib.style.core.USER_LIBRARY_PATHS.append(
    os.path.join(os.path.dirname(__file__), "stylelib")
)
matplotlib.style.core.reload_library()



from __future__ import annotations

__all__ = [
    "ItoolCrosshairControls",
    "ColorControls",
    "ItoolColormapControls",
    "ItoolBinningControls",
]

from typing import TYPE_CHECKING

import numpy as np
import pyqtgraph as pg
import qtawesome as qta
from qtpy import QtCore, QtGui, QtWidgets

from erlab.interactive.colors import ColorMapComboBox, ColorMapGammaWidget
from erlab.interactive.utilities import BetterSpinBox

if TYPE_CHECKING:
    import xarray as xr

    from erlab.interactive.imagetool.core import ImageSlicerArea
    from erlab.interactive.imagetool.slicer import ArraySlicer


class IconButton(QtWidgets.QPushButton):
    ICON_ALIASES = {
        "invert": "mdi6.invert-colors",
        "invert_off": "mdi6.invert-colors-off",
        "contrast": "mdi6.contrast-box",
        "lock": "mdi6.lock",
        "unlock": "mdi6.lock-open-variant",
        "bright_auto": "mdi6.brightness-auto",
        "bright_percent": "mdi6.brightness-percent",
        "colorbar": "mdi6.gradient-vertical",
        "transpose_0": "mdi6.arrow-top-left-bottom-right",
        "transpose_1": "mdi6.arrow-up-down",
        "transpose_2": "mdi6.arrow-left-right",
        "transpose_3": "mdi6.axis-z-arrow",
        "snap": "mdi6.grid",
        "snap_off": "mdi6.grid-off",
        "palette": "mdi6.palette-advanced",
        "styles": "mdi6.palette-swatch",
        "layout": "mdi6.page-layout-body",
        "zero_center": "mdi6.format-vertical-align-center",
        "table_eye": "mdi6.table-eye",
        "plus": "mdi6.plus",
        "minus": "mdi6.minus",
        "reset": "mdi6.backup-restore",
        # all_cursors="mdi6.checkbox-multiple-outline",
        "all_cursors": "mdi6.select-multiple",
    }

    def __init__(self, on: str | None = None, off: str | None = None, **kwargs):
        self.icon_key_on = None
        self.icon_key_off = None

        if on is not None:
            self.icon_key_on = on
            kwargs["icon"] = self.get_icon(self.icon_key_on)

        if off is not None:
            if on is None and kwargs["icon"] is None:
                raise ValueError("Icon for `on` state was not supplied.")
            self.icon_key_off = off
            kwargs.setdefault("checkable", True)

        super().__init__(**kwargs)
        if self.isCheckable() and off is not None:
            self.toggled.connect(self.refresh_icons)

    def setChecked(self, value: bool):
        super().setChecked(value)
        self.refresh_icons()

    def get_icon(self, icon: str):
        try:
            return qta.icon(self.ICON_ALIASES[icon])
        except KeyError:
            return qta.icon(icon)

    def refresh_icons(self):
        if self.icon_key_off is not None:
            if self.isChecked():
                self.setIcon(self.get_icon(self.icon_key_off))
                return
        if self.icon_key_on is not None:
            self.setIcon(self.get_icon(self.icon_key_on))

    def changeEvent(self, evt: QtCore.QEvent):  # handles dark mode
        if evt.type() == QtCore.QEvent.Type.PaletteChange:
            qta.reset_cache()
            self.refresh_icons()
        super().changeEvent(evt)


def clear_layout(layout: QtWidgets.QLayout):
    while layout.count():
        child = layout.takeAt(0)
        if child.widget():
            child.widget().deleteLater()


class ItoolControlsBase(QtWidgets.QWidget):
    def __init__(
        self, slicer_area: ImageSlicerArea | ItoolControlsBase, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._slicer_area = slicer_area
        self.sub_controls = []
        self.initialize_layout()
        self.initialize_widgets()
        self.connect_signals()
        self.update()

    @property
    def data(self) -> xr.DataArray:
        return self.slicer_area.data

    @property
    def array_slicer(self) -> ArraySlicer:
        return self.slicer_area.array_slicer

    @property
    def n_cursors(self) -> int:
        return self.slicer_area.n_cursors

    @property
    def current_cursor(self) -> int:
        return self.slicer_area.current_cursor

    def initialize_layout(self):
        self.setLayout(QtWidgets.QHBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.layout().setSpacing(3)

    def initialize_widgets(self):
        for ctrl in self.sub_controls:
            if isinstance(ctrl, ItoolControlsBase):
                ctrl.initialize_widgets()

    def connect_signals(self):
        for ctrl in self.sub_controls:
            if isinstance(ctrl, ItoolControlsBase):
                ctrl.connect_signals()

    def disconnect_signals(self):
        # Multiple inheritance disconnection is broken
        # https://bugreports.qt.io/browse/PYSIDE-229
        # Will not work correctly until this is fixed
        for ctrl in self.sub_controls:
            if isinstance(ctrl, ItoolControlsBase):
                ctrl.disconnect_signals()

    def update(self):
        for ctrl in self.sub_controls:
            ctrl.update()

    def add_control(self, widget: QtWidgets.QWidget) -> QtWidgets.QWidget:
        self.sub_controls.append(widget)
        return widget

    @property
    def is_nested(self) -> bool:
        return isinstance(self._slicer_area, ItoolControlsBase)

    @property
    def slicer_area(self) -> ImageSlicerArea:
        if isinstance(self._slicer_area, ItoolControlsBase):
            return self._slicer_area.slicer_area
        else:
            return self._slicer_area

    @slicer_area.setter
    def slicer_area(self, value: ImageSlicerArea):
        """
        Initially, the goal was to be able to control multiple `ImageSlicerArea`s
        with a single control widget, so the control widgets were designed with easy
        connection and disconnection of signals in mind. However, this has become
        largely unnecessary since we now have `SlicerLinkProxy` as a translation layer
        between multiple `ImageSlicerArea`s with individual control widgets. Hence, this
        method will remain unused for the time being.

        Also, in principle, most of the control widgets along with the menu bar should
        be re-written to use QActions...

        """

        # ignore until https://bugreports.qt.io/browse/PYSIDE-229 is fixed
        try:
            self.disconnect_signals()
        except RuntimeError:
            pass
        self._slicer_area = value
        clear_layout(self.layout())
        self.sub_controls = []
        self.initialize_widgets()
        self.update()
        self.connect_signals()

        print("called!")


# class ItoolAAAAAControls(ItoolControlsBase):
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)

#     def initialize_layout(self):
#         pass

#     def initialize_widgets(self):
#         pass

#     def connect_signals(self):
#         pass

#     def disconnect_signals(self):
#         pass

#     def update(self):
#         pass


class ItoolCrosshairControls(ItoolControlsBase):
    def __init__(self, *args, orientation=QtCore.Qt.Orientation.Vertical, **kwargs):
        if isinstance(orientation, QtCore.Qt.Orientation):
            self.orientation = orientation
        elif orientation == "vertical":
            self.orientation = QtCore.Qt.Orientation.Vertical
        elif orientation == "horizontal":
            self.orientation = QtCore.Qt.Orientation.Horizontal
        super().__init__(*args, **kwargs)

    def initialize_widgets(self):
        super().initialize_widgets()
        self.values_groups = tuple(
            QtWidgets.QWidget() for _ in range(self.data.ndim + 1)
        )
        self.values_layouts = tuple(
            QtWidgets.QGridLayout(g) for g in self.values_groups
        )
        for s in self.values_layouts:
            s.setContentsMargins(0, 0, 0, 0)

            s.setSpacing(3)
        # buttons for multicursor control
        self.btn_add = IconButton("plus", toolTip="Add cursor")
        self.btn_add.clicked.connect(self.slicer_area.add_cursor)

        self.btn_rem = IconButton("minus", toolTip="Remove cursor")
        self.btn_rem.clicked.connect(
            lambda: self.slicer_area.remove_cursor(self.cb_cursors.currentIndex())
        )

        self.btn_snap = IconButton(
            on="snap", off="snap_off", toolTip="Snap cursor to data points"
        )
        self.btn_snap.toggled.connect(self.slicer_area.toggle_snap)

        # multicursor combobox
        self.cb_cursors = QtWidgets.QComboBox()
        self.cb_cursors.textActivated.connect(self.setActiveCursor)
        self.cb_cursors.setMaximumHeight(
            QtGui.QFontMetrics(self.cb_cursors.font()).height() + 3
        )
        self.cb_cursors.setIconSize(QtCore.QSize(10, 10))
        for i in range(self.n_cursors):
            self.cb_cursors.addItem(self._cursor_icon(i), self._cursor_name(i))
        if self.n_cursors == 1:
            # can't remove more cursors
            self.cb_cursors.setDisabled(True)
            self.btn_rem.setDisabled(True)

        # current value widget
        self.spin_dat = BetterSpinBox(
            self.values_groups[-1], discrete=False, scientific=True, readOnly=True
        )
        try:
            with np.errstate(divide="ignore"):
                self.spin_dat.setDecimals(
                    round(abs(np.log10(self.array_slicer.absnanmax)) + 1)
                )
        except OverflowError:
            self.spin_dat.setDecimals(4)

        # add multicursor widgets
        if self.orientation == QtCore.Qt.Orientation.Vertical:
            self.values_layouts[0].addWidget(self.btn_add, 0, 1, 1, 1)
            self.values_layouts[0].addWidget(self.btn_rem, 0, 2, 1, 1)
            self.values_layouts[0].addWidget(self.btn_snap, 0, 0, 1, 1)
            self.values_layouts[0].addWidget(self.cb_cursors, 1, 0, 1, 3)
            self.values_layouts[0].addWidget(self.spin_dat, 2, 0, 1, 3)
        else:
            self.values_layouts[0].addWidget(self.btn_add, 0, 1, 1, 1)
            self.values_layouts[0].addWidget(self.btn_rem, 0, 2, 1, 1)
            self.values_layouts[0].addWidget(self.btn_snap, 0, 3, 1, 1)
            self.values_layouts[0].addWidget(self.cb_cursors, 0, 0, 1, 1)
            self.values_layouts[0].addWidget(self.spin_dat, 0, 4, 1, 1)
        self.layout().addWidget(self.values_groups[0])

        # info widgets
        self.label_dim = tuple(
            QtWidgets.QPushButton(grp, checkable=True) for grp in self.values_groups[1:]
        )
        self.spin_idx = tuple(
            BetterSpinBox(
                grp,
                integer=True,
                singleStep=1,
                wrapping=False,
                minimumWidth=60,
                keyboardTracking=False,
            )
            for grp in self.values_groups[1:]
        )
        self.spin_val = tuple(
            BetterSpinBox(
                grp,
                discrete=True,
                decimals=3,
                wrapping=False,
                minimumWidth=70,
                keyboardTracking=False,
            )
            for grp in self.values_groups[1:]
        )

        if self.data.ndim <= 3:
            icons = [f"transpose_{i}" for i in range(self.data.ndim)]
        else:
            icons = [f"transpose_{i}" for i in (0, 1, 3, 2)]
        self.btn_transpose = tuple(IconButton(on=i) for i in icons)

        # add and connect info widgets
        for i in range(self.data.ndim):
            # TODO: implelemnt cursor locking
            # self.label_dim[i].toggled.connect()
            self.spin_idx[i].valueChanged.connect(
                lambda ind, axis=i: self.slicer_area.set_index(axis, ind)
            )
            self.spin_val[i].valueChanged.connect(
                lambda val, axis=i: self.slicer_area.set_value(axis, val, uniform=False)
            )
            self.btn_transpose[i].clicked.connect(
                lambda *, idx=i: self._transpose_axes(idx)
            )
            if self.orientation == QtCore.Qt.Orientation.Vertical:
                self.values_layouts[i + 1].addWidget(self.label_dim[i], 0, 0, 1, 1)
                self.values_layouts[i + 1].addWidget(self.btn_transpose[i], 0, 1, 1, 1)
                self.values_layouts[i + 1].addWidget(self.spin_idx[i], 1, 0, 1, 2)
                self.values_layouts[i + 1].addWidget(self.spin_val[i], 2, 0, 1, 2)
            else:
                self.values_layouts[i + 1].addWidget(self.label_dim[i], 0, 0, 1, 1)
                self.values_layouts[i + 1].addWidget(self.btn_transpose[i], 0, 1, 1, 1)
                self.values_layouts[i + 1].addWidget(self.spin_idx[i], 0, 2, 1, 1)
                self.values_layouts[i + 1].addWidget(self.spin_val[i], 0, 3, 1, 1)

            self.layout().addWidget(self.values_groups[i + 1])

    def _transpose_axes(self, idx):
        if self.data.ndim == 4:
            if idx == 3:
                self.slicer_area.swap_axes(0, 2)
            else:
                self.slicer_area.swap_axes(idx, (idx + 1) % 4)
        else:
            self.slicer_area.swap_axes(idx, (idx + 1) % self.data.ndim)

    def connect_signals(self):
        super().connect_signals()
        self.slicer_area.sigDataChanged.connect(self.update)
        self.slicer_area.sigShapeChanged.connect(self.update)
        self.slicer_area.sigCurrentCursorChanged.connect(self.cursorChangeEvent)
        self.slicer_area.sigCursorCountChanged.connect(self.update_cursor_count)
        self.slicer_area.sigViewOptionChanged.connect(self.update_options)
        self.slicer_area.sigIndexChanged.connect(self.update_spins)
        self.slicer_area.sigBinChanged.connect(self.update_spins)

    def disconnect_signals(self):
        super().disconnect_signals()
        self.slicer_area.sigDataChanged.disconnect(self.update)
        self.slicer_area.sigShapeChanged.disconnect(self.update)
        self.slicer_area.sigCurrentCursorChanged.disconnect(self.cursorChangeEvent)
        self.slicer_area.sigViewOptionChanged.disconnect(self.update_options)
        self.slicer_area.sigCursorCountChanged.disconnect(self.update_cursor_count)
        self.slicer_area.sigIndexChanged.disconnect(self.update_spins)
        self.slicer_area.sigBinChanged.disconnect(self.update_spins)

    @QtCore.Slot()
    def update(self):
        super().update()
        if len(self.label_dim) != self.data.ndim:
            # number of required cursors changed, resetting
            clear_layout(self.layout())
            self.initialize_widgets()

        for i in range(self.data.ndim):
            self.values_groups[i].blockSignals(True)
            self.spin_idx[i].blockSignals(True)
            self.spin_val[i].blockSignals(True)

            if i in self.array_slicer._nonuniform_axes:
                self.label_dim[i].setText(str(self.data.dims[i])[:-4])
            else:
                self.label_dim[i].setText(str(self.data.dims[i]))

            lw = (
                self.label_dim[i]
                .fontMetrics()
                .boundingRect(self.label_dim[i].text())
                .width()
            )
            self.label_dim[i].setMaximumWidth(lw + 15)

            # update spinbox properties to match new data
            self.spin_idx[i].setRange(0, self.data.shape[i] - 1)
            self.spin_idx[i].setValue(self.slicer_area.get_current_index(i))

            self.spin_val[i].setRange(*self.array_slicer.lims[i])
            self.spin_val[i].setSingleStep(self.array_slicer.incs[i])
            self.spin_val[i].setValue(self.slicer_area.get_current_value(i))

            self.label_dim[i].blockSignals(False)
            self.spin_idx[i].blockSignals(False)
            self.spin_val[i].blockSignals(False)
        try:
            with np.errstate(divide="ignore"):
                self.spin_dat.setDecimals(
                    round(abs(np.log10(self.array_slicer.absnanmax)) + 1)
                )
        except OverflowError:
            self.spin_dat.setDecimals(4)
        self.spin_dat.setValue(
            self.array_slicer.point_value(self.current_cursor, binned=True)
        )

    def update_spins(self, *, axes=None):
        if axes is None:
            axes = range(self.data.ndim)
        if len(axes) != len(self.spin_idx):
            # called from sigIndexChanged before update was called
            return

        for i in axes:
            self.spin_idx[i].blockSignals(True)
            self.spin_val[i].blockSignals(True)
            self.spin_idx[i].setValue(self.slicer_area.current_indices[i])
            self.spin_val[i].setValue(self.slicer_area.current_values[i])
            self.spin_idx[i].blockSignals(False)
            self.spin_val[i].blockSignals(False)

        self.spin_dat.setValue(
            self.array_slicer.point_value(self.current_cursor, binned=True)
        )

    @QtCore.Slot()
    def update_options(self):
        self.btn_snap.blockSignals(True)
        self.btn_snap.setChecked(self.array_slicer.snap_to_data)
        # self.btn_snap.refresh_icons()
        self.btn_snap.blockSignals(False)

    def _cursor_name(self, i: int) -> str:
        # for cursor combobox content
        return f" Cursor {int(i)}"

    def _cursor_icon(self, i: int) -> QtGui.QIcon:
        img = QtGui.QImage(32, 32, QtGui.QImage.Format.Format_RGBA64)
        img.fill(QtCore.Qt.GlobalColor.transparent)

        painter = QtGui.QPainter(img)
        painter.setRenderHints(QtGui.QPainter.RenderHint.Antialiasing, True)

        clr = self.slicer_area.cursor_colors[i]
        painter.setBrush(pg.mkColor(clr))
        painter.drawEllipse(img.rect())
        painter.end()
        return QtGui.QIcon(QtGui.QPixmap.fromImage(img))

    @QtCore.Slot(int)
    def update_cursor_count(self, count: int):
        if count == self.cb_cursors.count():
            return
        elif count > self.cb_cursors.count():
            self.addCursor()
        else:
            self.remCursor()

    def addCursor(self):
        self.cb_cursors.setDisabled(False)
        # self.slicer_area.add_cursor()
        self.cb_cursors.addItem(
            self._cursor_icon(self.current_cursor),
            self._cursor_name(self.current_cursor),
        )
        self.cb_cursors.setCurrentIndex(self.current_cursor)
        self.btn_rem.setDisabled(False)

    def remCursor(self):
        # self.slicer_area.remove_cursor(self.cb_cursors.currentIndex())
        self.cb_cursors.removeItem(self.cb_cursors.currentIndex())
        for i in range(self.cb_cursors.count()):
            self.cb_cursors.setItemText(i, self._cursor_name(i))
            self.cb_cursors.setItemIcon(i, self._cursor_icon(i))
        self.cb_cursors.setCurrentText(self._cursor_name(self.current_cursor))
        if i == 0:
            self.cb_cursors.setDisabled(True)
            self.btn_rem.setDisabled(True)

    @QtCore.Slot(int)
    def cursorChangeEvent(self, idx: int):
        self.cb_cursors.setCurrentIndex(idx)
        self.update_spins()

    @QtCore.Slot(str)
    def setActiveCursor(self, value: str):
        self.slicer_area.set_current_cursor(self.cb_cursors.findText(value))


class ColorControls(ItoolControlsBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def initialize_widgets(self):
        self.btn_reverse = IconButton(
            on="invert",
            off="invert_off",
            checkable=True,
            toolTip="Invert colormap",
        )
        self.btn_contrast = IconButton(
            on="contrast",
            checkable=True,
            toolTip="High contrast mode",
        )
        self.btn_zero = IconButton(
            on="zero_center",
            checkable=True,
            toolTip="Keep center color fixed",
        )
        self.btn_lock = IconButton(
            # on="unlock", off="lock",
            on="bright_auto",
            off="bright_percent",
            checkable=True,
            toolTip="Lock color limits",
        )
        self.btn_reverse.toggled.connect(self.update_colormap)
        self.btn_contrast.toggled.connect(self.update_colormap)
        self.btn_zero.toggled.connect(self.update_colormap)

        self.layout().addWidget(self.btn_reverse)
        self.layout().addWidget(self.btn_contrast)
        self.layout().addWidget(self.btn_zero)
        self.layout().addWidget(self.btn_lock)

    def update(self):
        self.btn_reverse.blockSignals(True)
        self.btn_contrast.blockSignals(True)
        self.btn_zero.blockSignals(True)
        self.btn_lock.blockSignals(True)

        self.btn_reverse.setChecked(self.slicer_area.colormap_properties["reversed"])
        self.btn_contrast.setChecked(
            self.slicer_area.colormap_properties["highContrast"]
        )
        self.btn_zero.setChecked(self.slicer_area.colormap_properties["zeroCentered"])
        self.btn_lock.setChecked(self.slicer_area.color_locked)

        self.btn_reverse.blockSignals(False)
        self.btn_contrast.blockSignals(False)
        self.btn_zero.blockSignals(False)
        self.btn_lock.blockSignals(False)

    def update_colormap(self):
        self.slicer_area.set_colormap(
            reversed=self.btn_reverse.isChecked(),
            highContrast=self.btn_contrast.isChecked(),
            zeroCentered=self.btn_zero.isChecked(),
        )

    def connect_signals(self):
        super().connect_signals()
        self.btn_lock.toggled.connect(self.slicer_area.lock_levels)
        self.slicer_area.sigViewOptionChanged.connect(self.update)

    def disconnect_signals(self):
        super().disconnect_signals()
        self.btn_lock.toggled.disconnect(self.slicer_area.lock_levels)
        self.slicer_area.sigViewOptionChanged.disconnect(self.update)


class ItoolColormapControls(ItoolControlsBase):
    def __init__(self, *args, orientation=QtCore.Qt.Orientation.Vertical, **kwargs):
        if isinstance(orientation, QtCore.Qt.Orientation):
            self.orientation = orientation
        elif orientation == "vertical":
            self.orientation = QtCore.Qt.Orientation.Vertical
        elif orientation == "horizontal":
            self.orientation = QtCore.Qt.Orientation.Horizontal
        super().__init__(*args, **kwargs)

    def initialize_layout(self):
        if self.orientation == QtCore.Qt.Orientation.Vertical:
            self.setLayout(QtWidgets.QVBoxLayout(self))
        else:
            self.setLayout(QtWidgets.QHBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.layout().setSpacing(3)

    def initialize_widgets(self):
        super().initialize_widgets()
        self.cb_colormap = ColorMapComboBox(self, maximumWidth=175)
        self.cb_colormap.textActivated.connect(self.change_colormap)

        self.gamma_widget = ColorMapGammaWidget(spin_cls=BetterSpinBox)
        self.gamma_widget.valueChanged.connect(
            lambda g: self.slicer_area.set_colormap(gamma=g)
        )
        self.gamma_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed
        )

        self.misc_controls = self.add_control(ColorControls(self))

        self.layout().addWidget(self.cb_colormap)
        self.layout().addWidget(self.gamma_widget)
        self.layout().addWidget(self.misc_controls)

    def update(self):
        super().update()
        if isinstance(self.slicer_area.colormap, str):
            self.cb_colormap.setDefaultCmap(self.slicer_area.colormap)
        self.gamma_widget.blockSignals(True)
        self.gamma_widget.setValue(self.slicer_area.colormap_properties["gamma"])
        self.gamma_widget.blockSignals(False)

    def change_colormap(self, name):
        if name == self.cb_colormap.LOAD_ALL_TEXT:
            self.cb_colormap.load_all()
        else:
            self.slicer_area.set_colormap(name)

    def connect_signals(self):
        super().connect_signals()
        self.slicer_area.sigViewOptionChanged.connect(self.update)

    def disconnect_signals(self):
        super().disconnect_signals()
        self.slicer_area.sigViewOptionChanged.disconnect(self.update)


class ItoolBinningControls(ItoolControlsBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def initialize_layout(self):
        self.setLayout(QtWidgets.QGridLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.layout().setSpacing(3)

    def initialize_widgets(self):
        super().initialize_widgets()
        self.labels = tuple(QtWidgets.QLabel() for _ in range(self.data.ndim))
        self.spins = tuple(
            BetterSpinBox(
                self,
                integer=True,
                singleStep=2,
                # minimumWidth=60,
                value=1,
                minimum=1,
                maximum=self.data.shape[i],
                keyboardTracking=False,
            )
            for i in range(self.data.ndim)
        )
        for i, spin in enumerate(self.spins):
            spin.valueChanged.connect(lambda n, axis=i: self._update_bin(axis, n))

        self.reset_btn = IconButton("reset", toolTip="Reset bins")
        self.reset_btn.clicked.connect(self.reset)

        self.all_btn = IconButton(
            on="all_cursors",
            checkable=True,
            toolTip="Apply for all cursors",
        )

        for i in range(self.data.ndim):
            self.layout().addWidget(self.labels[i], 0, i, 1, 1)
            self.layout().addWidget(self.spins[i], 1, i, 1, 1)
        self.layout().addWidget(self.reset_btn, 2, 0, 1, 1)
        self.layout().addWidget(self.all_btn, 2, 1, 1, 1)
        # for spin in self.spins:
        # spin.setMinimumWidth(60)

    def _update_bin(self, axis, n):
        if self.all_btn.isChecked():
            self.slicer_area.set_bin_all(axis, n)
        else:
            self.slicer_area.set_bin(axis, n)

    def connect_signals(self):
        super().connect_signals()
        self.slicer_area.sigCurrentCursorChanged.connect(self.update)
        self.slicer_area.sigBinChanged.connect(self.update)
        self.slicer_area.sigDataChanged.connect(self.update)
        self.slicer_area.sigShapeChanged.connect(self.update)

    def disconnect_signals(self):
        super().disconnect_signals()
        self.slicer_area.sigCurrentCursorChanged.disconnect(self.update)
        self.slicer_area.sigBinChanged.disconnect(self.update)
        self.slicer_area.sigDataChanged.disconnect(self.update)
        self.slicer_area.sigShapeChanged.disconnect(self.update)

    def update(self):
        super().update()

        if len(self.labels) != self.data.ndim:
            clear_layout(self.layout())
            self.initialize_widgets()

        for i in range(self.data.ndim):
            self.spins[i].blockSignals(True)
            self.labels[i].setText(str(self.data.dims[i]))
            self.spins[i].setRange(1, self.data.shape[i] - 1)
            self.spins[i].setValue(self.array_slicer.get_bins(self.current_cursor)[i])
            self.spins[i].blockSignals(False)

    def reset(self):
        for spin in self.spins:
            spin.setValue(1)



import os

import h5netcdf
import numpy as np
import xarray as xr

import igor2.binarywave
import igor2.packed
import igor2.record

__all__ = ["load_experiment", "load_h5", "load_wave", "load_pxp", "load_ibw"]


def _load_experiment_raw(
    filename: str | os.PathLike,
    folder: str | None = None,
    *,
    prefix: str | None = None,
    ignore: list[str] | None = None,
    recursive: bool = False,
    **kwargs: dict,
) -> xr.Dataset:
    if folder is None:
        folder = []
    if ignore is None:
        ignore = []

    for bo in [">", "=", "<"]:
        try:
            _, expt = igor2.packed.load(filename, initial_byte_order=bo)
            break
        except ValueError:
            continue

    waves = {}
    if isinstance(folder, str):
        folder = folder.split("/")
    folder = [n.encode() for n in folder]

    expt = expt["root"]
    for dirname in folder:
        expt = expt[dirname]

    def unpack_folders(expt):
        for name, record in expt.items():
            if isinstance(record, igor2.record.WaveRecord):
                if prefix is not None:
                    if not name.decode().startswith(prefix):
                        continue
                if name.decode() in ignore:
                    continue
                waves[name.decode()] = load_wave(record, **kwargs)
            elif isinstance(record, dict):
                if recursive:
                    unpack_folders(record)

    unpack_folders(expt)
    return waves


def load_experiment(
    filename: str | os.PathLike,
    folder: str | None = None,
    *,
    prefix: str | None = None,
    ignore: list[str] | None = None,
    recursive: bool = False,
    **kwargs: dict,
) -> xr.Dataset:
    """Loads waves from an igor experiment (`.pxp`) file.

    Parameters
    ----------
    filename
        The experiment file.
    folder
        Target folder within the experiment, given as a slash-separated string. If
        `None`, defaults to the root.
    prefix
        If given, only include waves with names that starts with the given string.
    ignore
        List of wave names to ignore.
    recursive
        If `True`, includes waves in child directories.
    **kwargs
        Extra arguments to :func:`load_wave`.

    Returns
    -------
    xarray.Dataset
        Dataset containing the waves.

    """
    return xr.Dataset(
        _load_experiment_raw(
            filename,
            folder,
            prefix=prefix,
            ignore=ignore,
            recursive=recursive,
            **kwargs,
        )
    )


def load_h5(filename):
    ncf = h5netcdf.File(filename, mode="r", phony_dims="sort")
    ds = xr.open_dataset(xr.backends.H5NetCDFStore(ncf))
    for dv in ds.data_vars:
        wavescale = ds[dv].attrs["IGORWaveScaling"]
        ds = ds.assign_coords(
            {
                dim: wavescale[i + 1, 1]
                + wavescale[i + 1, 0] * np.arange(ds[dv].shape[i])
                for i, dim in enumerate(ds[dv].dims)
            }
        )
    return ds


def load_wave(
    wave: dict | igor2.record.WaveRecord | str | os.PathLike,
    data_dir: str | os.PathLike | None = None,
) -> xr.DataArray:
    """Load a wave from Igor binary format.

    Parameters
    ----------
    wave
        The wave to load. It can be provided as a dictionary, an instance of
        `igor2.record.WaveRecord`, or a string representing the path to the wave file.
    data_dir
        The directory where the wave file is located. This parameter is only used if
        `wave` is a string or `PathLike` object. If `None`, `wave` must be a valid path.
    Returns
    -------
    xarray.DataArray
        The loaded wave.

    Raises
    ------
    ValueError
        If the wave file cannot be found or loaded.
    TypeError
        If the wave argument is of an unsupported type.

    """
    DEFAULT_DIMS = ["W", "X", "Y", "Z"]
    _MAXDIM = 4

    if isinstance(wave, dict):
        wave_dict = wave
    elif isinstance(wave, igor2.record.WaveRecord):
        wave_dict = wave.wave
    else:
        if data_dir is not None:
            wave = os.path.join(data_dir, wave)
        wave_dict = igor2.binarywave.load(wave)

    d = wave_dict["wave"]
    version = wave_dict["version"]
    dim_labels = [""] * _MAXDIM
    bin_header, wave_header = d["bin_header"], d["wave_header"]
    if version <= 3:
        shape = [wave_header["npnts"]] + [0] * (_MAXDIM - 1)
        sfA = [wave_header["hsA"]] + [0] * (_MAXDIM - 1)
        sfB = [wave_header["hsB"]] + [0] * (_MAXDIM - 1)
        # data_units = wave_header["dataUnits"]
        axis_units = [wave_header["xUnits"]]
        axis_units.extend([""] * (_MAXDIM - len(axis_units)))
    else:
        shape = wave_header["nDim"]
        sfA = wave_header["sfA"]
        sfB = wave_header["sfB"]
        if version >= 5:
            # data_units = d["data_units"].decode()
            axis_units = [b"".join(d).decode() for d in wave_header["dimUnits"]]
            units_sizes = bin_header["dimEUnitsSize"]
            sz_cum = 0
            for i, sz in enumerate(units_sizes):
                if sz != 0:
                    axis_units[i] = d["dimension_units"][sz_cum : sz_cum + sz].decode()
                sz_cum += sz
            for i, sz in enumerate(bin_header["dimLabelsSize"]):
                if sz != 0:
                    dim_labels[i] = b"".join(d["labels"][i]).decode()
        else:
            # data_units = d["data_units"].decode()
            axis_units = [d["dimension_units"].decode()]

    def get_dim_name(index):
        dim = dim_labels[index]
        unit = axis_units[index]
        if dim == "":
            if unit == "":
                return DEFAULT_DIMS[index]
            else:
                return unit
        elif unit == "":
            return dim
        else:
            return f"{dim} ({unit})"

    dims = [get_dim_name(i) for i in range(_MAXDIM)]
    coords = {
        dims[i]: np.linspace(b, b + a * (c - 1), c)
        for i, (a, b, c) in enumerate(zip(sfA, sfB, shape))
        if c != 0
    }

    attrs = {}
    for ln in d.get("note", "").decode().splitlines():
        if "=" in ln:
            k, v = ln.split("=", 1)
            try:
                v = int(v)
            except ValueError:
                try:
                    v = float(v)
                except ValueError:
                    pass
            attrs[k] = v

    return xr.DataArray(
        d["wData"], dims=coords.keys(), coords=coords, attrs=attrs
    ).rename(wave_header["bname"].decode())


load_pxp = load_experiment
load_ibw = load_wave

# from arpes import load_pxt
# import igor.igorpy


# def load_pxp_old(filename, recursive=False, silent=False, **kwargs):
#     expt = load_pxt.read_experiment(filename, **kwargs)
#     waves = {}

#     def process_wave(arr):
#         arr = arr.where(arr != 0)
#         for d in arr.dims:
#             arr = arr.sortby(d)
#         return arr

#     def unpack_folders(expt):
#         for e in expt:
#             try:
#                 arr = process_wave(load_pxt.wave_to_xarray(e))
#                 waves[arr.name] = arr
#                 if not silent:
#                     print(arr.name)
#             except AttributeError:
#                 pass
#             if recursive and isinstance(e, igor.igorpy.Folder):
#                 unpack_folders(e)

#     unpack_folders(expt)
#     return waves


# def load_ibw_old(filename, data_dir: str | os.PathLike | None = None):
#     try:
#         filename = find_first_file(filename, data_dir=data_dir)
#     except (ValueError, TypeError):
#         if data_dir is not None:
#             filename = os.path.join(data_dir, filename)

#     class ibwfile_wave(object):
#         def __init__(self, fname):
#             self.wave = load_pxt.read_single_ibw(fname)

#     return load_pxt.wave_to_xarray(igor.igorpy.Wave(ibwfile_wave(filename)))



import sys

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors
from matplotlib.backends.backend_qtagg import FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.backends.qt_compat import QtCore, QtWidgets
from matplotlib.figure import Figure
from matplotlib.patches import Circle, Ellipse

from erlab.interactive.utilities import parse_data
from erlab.plotting.bz import plot_hex_bz
from erlab.plotting.colors import CenteredInversePowerNorm


def plot_bz_tise2(ax=None, a=3.54, pockets=False, aspect=1, rotate=0, **kwargs):
    """
    Plots a TiSe2 BZ on top of specified axis.
    """
    kwargs.setdefault("alpha", 1)
    kwargs.setdefault("color", "w")
    kwargs.setdefault("linestyle", "-")
    kwargs.setdefault("linewidth", 0.75)
    kwargs.setdefault("zorder", 5)
    # kwargs.setdefault('ax',plt.gca())

    # warnings.warn('Deprecated plot_bz_tise2')
    if ax is None:
        ax = plt.gca()
    plot_hex_bz(a=a, rotate=rotate, ax=ax)
    if pockets is True:
        color = kwargs.pop("color", None)
        kwargs["edgecolor"] = color
        kwargs["facecolor"] = "none"

        ln = 2 * np.pi / (a * 3)
        width = 0.25
        center = (0, np.sqrt(3) * ln)
        offset = (0, 0)
        for ang in [np.deg2rad(rotate - 30 + a) for a in [0, 60, 120, 180, 240, 300]]:
            x = np.cos(ang) * center[0] - np.sin(ang) * center[1]
            y = np.sin(ang) * center[0] + np.cos(ang) * center[1]
            if aspect == 1:
                p = Circle((x + offset[0], y + offset[1]), radius=width / 2, **kwargs)
            else:
                p = Ellipse(
                    (x + offset[0], y + offset[1]),
                    width=width,
                    height=width * aspect,
                    angle=np.rad2deg(ang),
                    **kwargs,
                )
            ax.add_patch(p)

        # p1 = Ellipse(
        #     (0, 2 * np.pi / (np.sqrt(3) * a)),
        #     width=0.2,
        #     height=0.2 * pocket_length,
        #     **kwargs
        # )
        # p2 = copy.deepcopy(p1)
        # p3 = copy.deepcopy(p1)
        # p4 = copy.deepcopy(p1)
        # p5 = copy.deepcopy(p1)
        # p6 = copy.deepcopy(p1)
        # p1.set_transform(Affine2D().rotate_deg(rotate - 30 + 0) + ax.transData)
        # p2.set_transform(Affine2D().rotate_deg(rotate - 30 + 60) + ax.transData)
        # p3.set_transform(Affine2D().rotate_deg(rotate - 30 + 120) + ax.transData)
        # p4.set_transform(Affine2D().rotate_deg(rotate - 30 + 180) + ax.transData)
        # p5.set_transform(Affine2D().rotate_deg(rotate - 30 + 240) + ax.transData)
        # p6.set_transform(Affine2D().rotate_deg(rotate - 30 + 300) + ax.transData)
        # ax.add_patch(p1)
        # ax.add_patch(p2)
        # ax.add_patch(p3)
        # ax.add_patch(p4)
        # ax.add_patch(p5)
        # ax.add_patch(p6)


class kTool(QtWidgets.QMainWindow):
    def __init__(
        self,
        data,
        bounds: dict | None = None,
        resolution: dict | None = None,
        gamma=0.5,
        cmap="twilight",
        plot_hex_bz=True,
        diff_data=None,
        a=3.54,
        rotate=90,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if bounds is None:
            bounds = {"kx": [-1.5, 1.5], "ky": [-1.5, 1.5], "kp": [-1.5, 1.5]}
        if resolution is None:
            resolution = {"kx": 0.02, "ky": 0.02}
        self.diff_data = diff_data
        self.data = parse_data(data)

        def _get_middle_index(x):
            return len(x) // 2 - (1 if len(x) % 2 == 0 else 0)

        self.has_eV = "eV" in self.data.dims
        self.show_eV = self.has_eV and self.data.ndim == 3
        if self.show_eV:
            self.data_all = self.data.copy(deep=True)
            self.coord_z = self.data_all["eV"].values
            self.ind_z = _get_middle_index(self.coord_z)
            self.data = self.data_all.isel(eV=self.ind_z)
            self.lims_z = (self.coord_z[0], self.coord_z[-1])
            self.inc_z = self.coord_z[1] - self.coord_z[0]
        self.dim_x, self.dim_y = self.data.dims
        self.coord_x, self.coord_y = (
            self.data[self.dim_x].values,
            self.data[self.dim_y].values,
        )
        self.lims_x, self.lims_y = (
            (self.coord_x[0], self.coord_x[-1]),
            (self.coord_y[0], self.coord_y[-1]),
        )
        self.vals = self.data.values
        self.bounds = bounds
        self.resolution = resolution
        self.kxy = self.data.kspace.convert(
            bounds=self.bounds, resolution=self.resolution
        )
        self.gamma = gamma
        self.cmap = cmap
        self.visible = True
        self.background = None
        improps = {
            "animated": True,
            "visible": True,
            "interpolation": "none",
            "aspect": "auto",
            "origin": "lower",
            "norm": colors.PowerNorm(self.gamma),
            "cmap": self.cmap,
        }

        self._main = QtWidgets.QWidget()
        self.setCentralWidget(self._main)
        self.layout = QtWidgets.QVBoxLayout(self._main)

        self.new_offsets_rad = {}
        self.new_offsets_deg = {}

        self.canvas = FigureCanvas(
            Figure(figsize=(8, 16), dpi=100, layout="constrained")
        )
        self.addToolBar(
            QtCore.Qt.BottomToolBarArea, NavigationToolbar(self.canvas, self)
        )

        gs = self.canvas.figure.add_gridspec(1, 2)
        self.ax0 = self.canvas.figure.add_subplot(gs[0])
        self.ax1 = self.canvas.figure.add_subplot(gs[1])
        self.ax0.grid()
        self.ax1.grid()
        self.ax0.set_aspect("equal")
        self.ax1.set_aspect("equal")

        self.im_r = self.ax0.imshow(
            self.vals, extent=(*self.lims_x, *self.lims_y), **improps
        )

        self.dispmode = QtWidgets.QComboBox()
        self.dispmode.addItems(["diff", "converted", "reference"])
        self.dispmode.setCurrentIndex(0)
        self.dispmode.currentIndexChanged.connect(self._update_all)

        self.im_k = self.ax1.imshow(self.disp_values, extent=self.extent, **improps)
        self.ax0.set_xlabel(self.dim_x)
        self.ax0.set_ylabel(self.dim_y)
        self.ax1.set_xlabel(self.dim_kx)
        self.ax1.set_ylabel(self.dim_ky)
        self.ax1.axline(
            (0, 0), (np.pi / 3.54, -np.pi / np.sqrt(3) / 3.54), ls="--", lw=0.75, c="w"
        )
        self.ax1.axline(
            (0, 0), (-np.pi / 3.54, -np.pi / np.sqrt(3) / 3.54), ls="--", lw=0.75, c="w"
        )

        self.offsetpanel = QtWidgets.QWidget()
        offsetpanelcontent = QtWidgets.QHBoxLayout()

        self.offsetcoords = ["theta", "phi", "beta", "psi", "chi"]

        spin0label = QtWidgets.QLabel(self.offsetcoords[0])
        spin1label = QtWidgets.QLabel(self.offsetcoords[1])
        spin2label = QtWidgets.QLabel(self.offsetcoords[2])
        spin3label = QtWidgets.QLabel(self.offsetcoords[3])
        spin4label = QtWidgets.QLabel(self.offsetcoords[4])

        self.spin0 = QtWidgets.QDoubleSpinBox()
        self.spin1 = QtWidgets.QDoubleSpinBox()
        self.spin2 = QtWidgets.QDoubleSpinBox()
        self.spin3 = QtWidgets.QDoubleSpinBox()
        self.spin4 = QtWidgets.QDoubleSpinBox()
        self.spin0.setRange(-180, 180)
        self.spin1.setRange(-180, 180)
        self.spin2.setRange(-180, 180)
        self.spin3.setRange(-180, 180)
        self.spin4.setRange(-180, 180)
        self.spin0.setSingleStep(0.01)
        self.spin1.setSingleStep(0.01)
        self.spin2.setSingleStep(0.01)
        self.spin3.setSingleStep(0.01)
        self.spin4.setSingleStep(0.01)
        self.spin0.setDecimals(3)
        self.spin1.setDecimals(3)
        self.spin2.setDecimals(3)
        self.spin3.setDecimals(3)
        self.spin4.setDecimals(3)
        self.spin0.setValue(0.0)
        self.spin1.setValue(0.0)
        self.spin2.setValue(0.0)
        self.spin3.setValue(0.0)
        self.spin4.setValue(0.0)
        self.spin0.valueChanged.connect(lambda v: self._spinchanged(0, v))
        self.spin1.valueChanged.connect(lambda v: self._spinchanged(1, v))
        self.spin2.valueChanged.connect(lambda v: self._spinchanged(2, v))
        self.spin3.valueChanged.connect(lambda v: self._spinchanged(3, v))
        self.spin4.valueChanged.connect(lambda v: self._spinchanged(4, v))

        offsetpanelcontent.addWidget(spin0label)
        offsetpanelcontent.addWidget(self.spin0)
        offsetpanelcontent.addWidget(spin1label)
        offsetpanelcontent.addWidget(self.spin1)
        offsetpanelcontent.addWidget(spin2label)
        offsetpanelcontent.addWidget(self.spin2)
        offsetpanelcontent.addWidget(spin3label)
        offsetpanelcontent.addWidget(self.spin3)
        offsetpanelcontent.addWidget(spin4label)
        offsetpanelcontent.addWidget(self.spin4)
        offsetpanelcontent.addStretch()
        self.offsetpanel.setLayout(offsetpanelcontent)

        self.colorstab = QtWidgets.QWidget()
        colorstabcontent = QtWidgets.QHBoxLayout()
        gammalabel = QtWidgets.QLabel("g")
        gammaspin = QtWidgets.QDoubleSpinBox()
        gammaspin.setToolTip("Colormap Gamma")
        gammaspin.setSingleStep(0.05)
        gammaspin.setValue(0.5)
        gammaspin.setMinimum(0.00001)
        gammaspin.valueChanged.connect(self._set_gamma)
        colormaps = QtWidgets.QComboBox()
        colormaps.setToolTip("Colormap")
        colormaps.addItems(plt.colormaps())
        colormaps.setCurrentIndex(colormaps.findText(self.cmap))
        colormaps.currentTextChanged.connect(self._set_cmap)
        colorstabcontent.addWidget(gammalabel)
        colorstabcontent.addWidget(gammaspin)
        colorstabcontent.addWidget(colormaps)
        colorstabcontent.addStretch()
        self.colorstab.setLayout(colorstabcontent)

        self.boundstab = QtWidgets.QWidget()
        boundstabcontent = QtWidgets.QHBoxLayout(self.boundstab)
        self.boundsgroup = QtWidgets.QGroupBox("Bounds")
        self.boundsgrid = QtWidgets.QGridLayout(self.boundsgroup)
        self.kxminlabel = QtWidgets.QLabel(self.boundsgroup)
        self.boundsgrid.addWidget(self.kxminlabel, 0, 0, 1, 1)
        self.kxminspin = QtWidgets.QDoubleSpinBox(self.boundsgroup)
        self.boundsgrid.addWidget(self.kxminspin, 0, 1, 1, 1)
        self.kxmaxlabel = QtWidgets.QLabel(self.boundsgroup)
        self.boundsgrid.addWidget(self.kxmaxlabel, 0, 2, 1, 1, QtCore.Qt.AlignHCenter)
        self.kxmaxspin = QtWidgets.QDoubleSpinBox(self.boundsgroup)
        self.boundsgrid.addWidget(self.kxmaxspin, 0, 3, 1, 1)
        self.kyminlabel = QtWidgets.QLabel(self.boundsgroup)
        self.boundsgrid.addWidget(self.kyminlabel, 1, 0, 1, 1)
        self.kyminspin = QtWidgets.QDoubleSpinBox(self.boundsgroup)
        self.boundsgrid.addWidget(self.kyminspin, 1, 1, 1, 1)
        self.kymaxlabel = QtWidgets.QLabel(self.boundsgroup)
        self.boundsgrid.addWidget(self.kymaxlabel, 1, 2, 1, 1, QtCore.Qt.AlignHCenter)
        self.kymaxspin = QtWidgets.QDoubleSpinBox(self.boundsgroup)
        self.boundsgrid.addWidget(self.kymaxspin, 1, 3, 1, 1)
        boundstabcontent.addWidget(self.boundsgroup)
        self.resolutiongroup = QtWidgets.QGroupBox("Resolution")
        self.resgrid = QtWidgets.QGridLayout(self.resolutiongroup)
        self.kxreslabel = QtWidgets.QLabel(self.resolutiongroup)
        self.resgrid.addWidget(self.kxreslabel, 0, 0, 1, 1)
        self.kxresspin = QtWidgets.QDoubleSpinBox(self.resolutiongroup)
        self.resgrid.addWidget(self.kxresspin, 0, 1, 1, 1)
        self.kyreslabel = QtWidgets.QLabel(self.resolutiongroup)
        self.resgrid.addWidget(self.kyreslabel, 1, 0, 1, 1)
        self.kyresspin = QtWidgets.QDoubleSpinBox(self.resolutiongroup)
        self.resgrid.addWidget(self.kyresspin, 1, 1, 1, 1)
        boundstabcontent.addWidget(self.resolutiongroup)
        self.kxminlabel.setText("kx")
        self.kxmaxlabel.setText("to")
        self.kyminlabel.setText("ky")
        self.kymaxlabel.setText("to")
        self.kxreslabel.setText("kx")
        self.kyreslabel.setText("ky")
        self.kxminspin.setSingleStep(0.05)
        self.kxmaxspin.setSingleStep(0.05)
        self.kyminspin.setSingleStep(0.05)
        self.kymaxspin.setSingleStep(0.05)
        self.kxminspin.setRange(-50, 50)
        self.kxmaxspin.setRange(-50, 50)
        self.kyminspin.setRange(-50, 50)
        self.kymaxspin.setRange(-50, 50)
        self.kxminspin.setValue(self.bounds["kx"][0])
        self.kxmaxspin.setValue(self.bounds["kx"][1])
        self.kyminspin.setValue(self.bounds["ky"][0])
        self.kymaxspin.setValue(self.bounds["ky"][1])
        self.kxminspin.valueChanged.connect(lambda v: self._set_bounds("kx", 0, v))
        self.kxmaxspin.valueChanged.connect(lambda v: self._set_bounds("kx", 1, v))
        self.kyminspin.valueChanged.connect(lambda v: self._set_bounds("ky", 0, v))
        self.kymaxspin.valueChanged.connect(lambda v: self._set_bounds("ky", 1, v))
        self.kxresspin.setValue(self.resolution["kx"])
        self.kyresspin.setValue(self.resolution["ky"])
        self.kxresspin.setSingleStep(0.001)
        self.kyresspin.setSingleStep(0.001)
        self.kxresspin.setMinimum(0.001)
        self.kyresspin.setMinimum(0.001)
        self.kxresspin.valueChanged.connect(lambda v: self._set_resolution("kx", v))
        self.kyresspin.valueChanged.connect(lambda v: self._set_resolution("ky", v))
        self.kxresspin.setDecimals(3)
        self.kyresspin.setDecimals(3)

        if self.diff_data is not None:
            boundstabcontent.addWidget(self.dispmode)

        self.tabwidget = QtWidgets.QTabWidget()
        self.tabwidget.addTab(self.offsetpanel, "Offsets")
        self.tabwidget.addTab(self.boundstab, "Bounds")
        self.tabwidget.addTab(self.colorstab, "Colors")
        self.tabwidget.setSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Maximum
        )
        self.canvas.setSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Expanding
        )

        self.layout.addWidget(self.tabwidget)
        if self.show_eV:
            zvaluepanel = QtWidgets.QWidget()
            zvaluecontent = QtWidgets.QHBoxLayout(zvaluepanel)
            self.zspin = QtWidgets.QDoubleSpinBox()
            self.zspin.setSingleStep(self.inc_z)
            self.zspin.setRange(*self.lims_z)
            self.zspin.setValue(self.coord_z[self.ind_z])
            self.zspin.valueChanged.connect(self._zspinchanged)
            self.zslider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
            self.zslider.setSingleStep(1)
            self.zslider.setPageStep(10)
            self.zslider.setMinimum(0)
            self.zslider.setMaximum(len(self.coord_z) - 1)
            self.zslider.setValue(self.ind_z)
            self.zslider.valueChanged.connect(self._zsliderchanged)
            zvaluecontent.addWidget(self.zspin)
            zvaluecontent.addWidget(self.zslider)
            self.layout.addWidget(zvaluepanel)

        self.layout.addWidget(self.canvas)
        # self.canvas.mpl_connect('draw_event', self.clear)
        if plot_hex_bz:
            plot_bz_tise2(a=a, ax=self.ax1, rotate=rotate, pockets=True)
        self.canvas.draw()
        self.spin4.setValue(np.rad2deg(data.chi))
        # self.im_r.set_visible(self.visible)
        # self.im_k.set_visible(self.visible)

    def clear(self, event):
        self.im_r.set_visible(False)
        self.im_k.set_visible(False)
        # self.background = self.canvas.copy_from_bbox(self.canvas.figure.bbox)

    def closeEvent(self, event):
        # do stuff
        print("ktool closed with offsets")
        print("in radians:")
        print(self.new_offsets_rad)
        print("in degrees:")
        print(self.new_offsets_deg)
        event.accept()

    def _update_plots(self):
        self.im_r.set_visible(self.visible)
        self.im_k.set_visible(self.visible)
        # if self.background is not None:
        # self.canvas.restore_region(self.background)
        self.ax0.draw_artist(self.im_r)
        self.ax1.draw_artist(self.im_k)
        # self.canvas.blit()
        self.canvas.draw()

    def _set_resolution(self, ax, value):
        self.resolution[ax] = value
        self._update_kxy()
        self._update_plots()

    def _update_all(self):
        self._update_disp()
        self._update_plots()

    def _set_bounds(self, ax, ind, value):
        self.bounds[ax][ind] = value
        self.bounds["kp"][ind] = value
        self._update_kxy()
        self._update_plots()
        if (ax == "kx") & (ind == 0):
            self.kxmaxspin.setMinimum(value)
        elif (ax == "kx") & (ind == 1):
            self.kxminspin.setMaximum(value)
        elif (ax == "ky") & (ind == 0):
            self.kymaxspin.setMinimum(value)
        elif (ax == "ky") & (ind == 1):
            self.kyminspin.setMaximum(value)

    def _zspinchanged(self, value):
        self.ind_z = np.rint((value - self.lims_z[0]) / self.inc_z).astype(int)
        self.zslider.blockSignals(True)
        self.zslider.setValue(self.ind_z)
        self.zslider.blockSignals(False)
        self._update_data()
        self._update_kxy()
        self.im_r.set_norm(colors.PowerNorm(self.gamma))
        self._update_plots()

    def _zsliderchanged(self, value):
        self.ind_z = value
        self.zspin.blockSignals(True)
        self.zspin.setValue(self.coord_z[self.ind_z])
        self.zspin.blockSignals(False)
        self._update_data()
        self._update_kxy()
        self._update_plots()

    def _update_data(self):
        self.data = self.data_all.isel(eV=self.ind_z)
        self.vals = self.data.values
        self.im_r.set_data(self.vals)
        self.im_r.set_norm(colors.PowerNorm(self.gamma))

    def _set_gamma(self, gamma):
        self.gamma = gamma
        self.im_r.set_norm(colors.PowerNorm(self.gamma))
        if self.diff_data is not None:
            self.im_k.set_norm(CenteredInversePowerNorm(self.gamma))
        else:
            self.im_k.set_norm(colors.PowerNorm(self.gamma))
        self._update_plots()

    def _set_cmap(self, cmap):
        self.cmap = cmap
        self.im_r.set_cmap(self.cmap)
        self.im_k.set_cmap(self.cmap)
        self._update_plots()

    def _spinchanged(self, n, value):
        # self.data.S.apply_offsets({self.offsetcoords[n]:value*np.pi/180})
        self.new_offsets_rad[self.offsetcoords[n]] = np.deg2rad(value)
        self.new_offsets_deg[self.offsetcoords[n]] = np.around(value, 3)
        self.data.S.apply_offsets(self.new_offsets_rad)
        if self.show_eV:
            self.data_all.S.apply_offsets(self.new_offsets_rad)
        self._update_kxy()
        self._update_plots()

    @property
    def disp_values(self):
        self.dim_ky, self.dim_kx = self.kxy.dims
        self.coord_kx, self.coord_ky = (
            self.kxy[self.dim_kx].values,
            self.kxy[self.dim_ky].values,
        )
        self.lims_kx, self.lims_ky = (
            (self.coord_kx[0], self.coord_kx[-1]),
            (self.coord_ky[0], self.coord_ky[-1]),
        )
        if self.diff_data is not None:
            try:
                ref = (
                    self.diff_data.sel(
                        eV=self.data_all.eV[self.ind_z], method="nearest"
                    )
                    .sel(kx=slice(*self.lims_kx), ky=slice(*self.lims_ky))
                    .values
                )
                match self.dispmode.currentText():
                    case "diff":
                        return self.kxy.values - ref
                    case "converted":
                        return self.kxy.values
                    case _:
                        return ref
            except ValueError:
                pass
        return self.kxy.values

    @property
    def extent(self):
        return (*self.lims_kx, *self.lims_ky)

    def _update_disp(self):
        self.im_k.set_data(self.disp_values)
        self.im_k.set_extent(self.extent)
        if self.diff_data is not None:
            self.im_k.set_norm(CenteredInversePowerNorm(self.gamma))
        else:
            self.im_k.set_norm(colors.PowerNorm(self.gamma))

    def _update_kxy(self):
        self.kxy = self.data.kspace.convert(
            bounds=self.bounds, resolution=self.resolution
        )
        self._update_disp()


def ktool(data, execute=True, *args, **kwargs):
    qapp = QtWidgets.QApplication.instance()
    if not qapp:
        qapp = QtWidgets.QApplication(sys.argv)
    with plt.rc_context(
        {
            "text.usetex": False,
            #     #  'mathtext.fontset':'stixsans',
            "font.size": 7,
            "font.family": "sans",
        }
    ):
        win = kTool(data, *args, **kwargs)
        win.show()
        win.activateWindow()
        win.raise_()
    if execute is None:
        execute = True
        try:
            shell = get_ipython().__class__.__name__  # type: ignore
            if shell == "ZMQInteractiveShell" or shell == "TerminalInteractiveShell":
                execute = False
        except NameError:
            pass
    if execute:
        qapp.exec()
    qapp.setStyle("Fusion")
    return win



from collections.abc import Callable

import erlab.analysis.kspace
import numpy as np
import pytest
import xarray
from erlab.interactive.exampledata import generate_data_angles

k_tot = np.array([1.0, 2.0, 3.0])


def _generate_funclist() -> list[tuple[Callable, Callable]]:
    funcs = []

    for kconv_func in (
        erlab.analysis.kspace._kconv_func_type1,
        erlab.analysis.kspace._kconv_func_type2,
    ):
        for delta, xi, xi0, beta0 in zip(
            [0, 90.0, -90.0],
            [0, 30.0, -30.0],
            [0.0, 10.0, -10.0],
            [0.0, 10.0, -10.0],
        ):
            funcs.append(kconv_func(k_tot, delta, xi, xi0, beta0))
    for kconv_func in (
        erlab.analysis.kspace._kconv_func_type1_da,
        erlab.analysis.kspace._kconv_func_type2_da,
    ):
        for delta, chi, chi0, xi, xi0 in zip(
            [0, 90.0, -90.0],
            [0, 10.0, -30.0],
            [0.0, 10.0, -10.0],
            [0.0, 10.0, -10.0],
            [0.0, 10.0, -10.0],
        ):
            funcs.append(kconv_func(k_tot, delta, chi, chi0, xi, xi0))
    return funcs


def test_transform():
    for forward_func, inverse_func in _generate_funclist():
        alpha, beta = inverse_func(*forward_func(10.0, 20.0))
        assert alpha.size == beta.size == k_tot.size
        assert np.allclose(alpha, 10.0)
        assert np.allclose(beta, 20.0)

        kx, ky = forward_func(*inverse_func(0.1, 0.2))
        assert kx.size == ky.size == k_tot.size
        assert np.allclose(kx, 0.1)
        assert np.allclose(ky, 0.2)


@pytest.fixture
def angle_data():
    return generate_data_angles(shape=(10, 10, 10))


def test_offsets(angle_data):
    angle_data.kspace.offsets.reset()
    angle_data.kspace.offsets = {"xi": 10.0}
    answer = {k: 0.0 for k in angle_data.kspace.valid_offset_keys}
    answer["xi"] = 10.0
    assert dict(angle_data.kspace.offsets) == answer


def test_kconv(angle_data):
    for conf in erlab.analysis.kspace.AxesConfiguration:
        angle_data.kspace.configuration = conf
        match conf:
            case (
                erlab.analysis.kspace.AxesConfiguration.Type1DA
                | erlab.analysis.kspace.AxesConfiguration.Type2DA
            ):
                angle_data = angle_data.assign_coords(chi=0.0)

        kconv = angle_data.kspace.convert()

        assert isinstance(kconv, xarray.DataArray)
        assert not kconv.isnull().all()



"""Some general functions and utilities used in fitting.

Many functions are `numba`-compiled for speed.
"""

__all__ = [
    "TINY",
    "do_convolve",
    "do_convolve_y",
    "gaussian_wh",
    "lorentzian_wh",
    "fermi_dirac",
    "fermi_dirac_linbkg",
    "fermi_dirac_linbkg_broad",
    "step_linbkg_broad",
    "step_broad",
]

from collections.abc import Callable

import numba
import numpy as np
import numpy.typing as npt
import scipy.special

from erlab.constants import kb_eV

#: From :mod:`lmfit.lineshapes`, equal to `numpy.finfo(numpy.float64).resolution`
TINY: float = 1.0e-15


@numba.njit(cache=True)
def _gen_kernel(
    x: npt.NDArray[np.float64], resolution: float, pad: int = 5
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    r"""Generate a Gaussian kernel for convolution.

    Parameters
    ----------
    x
        The input array of x values.
    resolution
        The resolution of the kernel given as FWHM.
    pad
        Multiples of the standard deviation :math:`\sigma` to truncate the kernel at.

    Returns
    -------
    extended
        The domain of the kernel.
    gauss
        The gaussian kernel defined on `extended`.

    """
    delta_x = x[1] - x[0]
    sigma = abs(resolution) / np.sqrt(8 * np.log(2))  # resolution given in FWHM
    n_pad = int(sigma * pad / delta_x + 0.5)
    x_pad = n_pad * delta_x

    extended = np.linspace(x[0] - x_pad, x[-1] + x_pad, 2 * n_pad + len(x))
    gauss = (
        delta_x
        * np.exp(
            -(np.linspace(-x_pad, x_pad, 2 * n_pad + 1) ** 2) / max(TINY, 2 * sigma**2)
        )
        / max(TINY, np.sqrt(2 * np.pi) * sigma)
    )
    return extended, gauss


def do_convolve(
    x: npt.NDArray[np.float64],
    func: Callable,
    resolution: float,
    pad: int = 5,
    **kwargs: dict,
) -> npt.NDArray[np.float64]:
    r"""Convolves `func` with gaussian of FWHM `resolution` in `x`.

    Parameters
    ----------
    x
        A evenly spaced array specifing where to evaluate the convolution.
    func
        Function to convolve.
    resolution
        FWHM of the gaussian kernel.
    pad
        Multiples of the standard deviation :math:`\sigma` to pad with.
    **kwargs
        Additional keyword arguments to `func`.

    """
    xn, g = _gen_kernel(
        np.asarray(x, dtype=np.float64), float(resolution), pad=int(pad)
    )
    return np.convolve(func(xn, **kwargs), g, mode="valid")


def do_convolve_y(
    x: npt.NDArray[np.float64],
    y: npt.NDArray[np.float64],
    func: Callable,
    resolution: float,
    pad: int = 5,
    **kwargs: dict,
) -> npt.NDArray[np.float64]:
    xn, g = _gen_kernel(
        np.asarray(np.squeeze(x), dtype=np.float64), resolution, pad=pad
    )
    if not np.iterable(y):
        y = [y]
    return np.vstack(
        [
            np.convolve(func(xn, yi, **kwargs), g, mode="valid")
            for yi in np.asarray(y).flat
        ]
    ).T


@numba.njit(cache=True)
def gaussian_wh(
    x: npt.NDArray[np.float64], center: float, width: float, height: float
) -> npt.NDArray[np.float64]:
    r"""Gaussian parametrized with FWHM and peak height.

    Note
    ----
    :math:`\sigma=\frac{w}{2\sqrt{2\log{2}}}`

    """
    return height * np.exp(
        -16 * np.log(2) * (1.0 * x - center) ** 2 / max(TINY, width**2)
    )


@numba.njit(cache=True)
def lorentzian_wh(
    x: npt.NDArray[np.float64], center: float, width: float, height: float
) -> npt.NDArray[np.float64]:
    r"""Lorentzian parametrized with FWHM and peak height.

    Note
    ----
    :math:`\sigma=w/2`

    """
    return height / (1 + 4 * ((1.0 * x - center) / max(TINY, width)) ** 2)


@numba.njit(cache=True)
def fermi_dirac(
    x: npt.NDArray[np.float64], center: float, temp: float
) -> npt.NDArray[np.float64]:
    """Fermi-dirac edge in terms of temperature."""
    return 1 / (1 + np.exp((1.0 * x - center) / max(TINY, temp * kb_eV)))


# adapted and improved from KWAN Igor procedures
@numba.njit(cache=True)
def fermi_dirac_linbkg(
    x: npt.NDArray[np.float64],
    center: float,
    temp: float,
    back0: float,
    back1: float,
    dos0: float,
    dos1: float,
) -> npt.NDArray[np.float64]:
    """Fermi-dirac edge with linear backgrounds above and below the fermi level.

    Note
    ----
    `back0` and `back1` corresponds to the linear background above and below EF (due to
    non-homogeneous detector efficiency or residual intensity on the phosphor screen
    during sweep mode), while `dos0` and `dos1` corresponds to the linear density of
    states below EF including the linear background.
    """
    return (back0 + back1 * x) + (dos0 - back0 + (dos1 - back1) * x) / (
        1 + np.exp((1.0 * x - center) / max(TINY, temp * kb_eV))
    )


def fermi_dirac_linbkg_broad(
    x: npt.NDArray[np.float64],
    center: float,
    temp: float,
    resolution: float,
    back0: float,
    back1: float,
    dos0: float,
    dos1: float,
) -> npt.NDArray[np.float64]:
    """
    Resolution-broadened Fermi-dirac edge with linear backgrounds above and below the
    fermi level.
    """
    return do_convolve(
        x,
        fermi_dirac_linbkg,
        resolution=resolution,
        center=center,
        temp=temp,
        back0=back0,
        back1=back1,
        dos0=dos0,
        dos1=dos1,
    )


def step_broad(
    x: npt.NDArray[np.float64],
    center: float = 0.0,
    sigma: float = 1.0,
    amplitude: float = 1.0,
):
    """Step function convolved with a Gaussian."""
    return (
        amplitude
        * 0.5
        * scipy.special.erfc((1.0 * x - center) / max(TINY, np.sqrt(2) * sigma))
    )


def step_linbkg_broad(
    x: npt.NDArray[np.float64],
    center: float,
    sigma: float,
    back0: float,
    back1: float,
    dos0: float,
    dos1: float,
):
    """
    A linear density of states multiplied with a resolution broadened step function with
    a linear background.
    """
    return (back0 + back1 * x) + (dos0 - back0 + (dos1 - back1) * x) * (
        step_broad(x, center, sigma, 1.0)
    )
